program InternetConfigMain;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	This module implements the main entry point of the Internet Config application.
	
	The rest of the file is made up of code that simply initialises the rest of the program.

	Useful Global Knowledge
	----------------------
	
	1. The source consistently uses certain prefixes to identify certain classes of
		identifiers. These include:

			kXxx		-- constant
			gXxx		-- global variable (used even for globals internal to individual modules)
			mXxx		-- menu ID
			iXxx		-- menu item number
			rXxx		-- resource ID
			strXxx		-- index into a STR# resource
			ditXxx		-- dialog item number
			WT_Xxx		-- Window Type, see ICWindowGlobals for a description
			acXxx		-- Error actions, see ICGlobals
			
		Also, the source follows the usual Macintosh naming conventions for routines
		(in MixedCase) and parameters/locals (withLowerCaseFirstLetter).
	
	2. Every now and again I turn on Case Sensitive Identifiers in Metrowerks
		Pascal to root out all the places where I've used the wrong case.
		
	3. mention memory management issues, specifically the fact that
		I don't worry about NewUPP failing because all of them are done very
		early.
*)

	uses
		Processes,
		Fonts,
		TextEdit,
		Dialogs,
		GestaltEqu,
		Traps,
		Appearance,

		InternetConfig,
		
		ICDebug,
		ICText,
		ICDialogs,
		ICStandardFile,
		ICResources,
		ICGlobals, 
		ICDocument,
		ICEvents, 
		ICWindows, 
		ICInstall, 
		ICIconSuites, 
		ICCommonSubs,
		ICMiscSubs, 
		ICMiscAppleEvents,
		ICPreferenceCache;

	function HandleAEQuitApplication (var theAppleEvent, reply: AppleEvent; refcon : UInt32): OSErr;
		(* Handle the Quit AppleEvent. Basically this does the standard AppleEvent stuff, then calls
			the DoQuit routine in ICDocuments. If this returns noErr, we set the gQuitNow Boolean
			which causes the application to exit the main event loop.
		*)
		var
			err: OSStatus;
	begin
		{$unused reply}
		{$unused refcon}
		err := AEGotRequiredParams(theAppleEvent);
		if err = noErr then begin
			err := DoQuit;
		end; (* if *)
		HandleAEQuitApplication := err;
	end; (* HandleAEQuitApplication *)

	procedure InstallAppleEventHandlers;
		(* This routine is called at application startup time. It's responsible for either registering the
			AppleEvent handlers in preparation for the main event loop getting the initial AppleEvents,
			or for reading through the AppParameters list and doing the right thing.
			Well it used to do both.  Now it just has System 7 code.
		*)
		var
			junk: OSStatus;
	begin
		junk := AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, 
								NewAEEventHandlerUPP(@HandleAEQuitApplication), 0, false);
		ICAssert(junk = noErr);
		InstallMiscAppleEventHandlers;
	end; (* InstallAppleEventHandlers *)

{$IFC not TARGET_RT_MAC_CFM}
	function Get68KStackPointerRegister: longint;
	inline
		$2E8F;				(* 	move.l a7,(sp) 	*)
{$ENDC}

	procedure Initialization;
		(* The canonical Toolbox initialisation stuff. Why is this canonical?  Because this is the sequence that
			Think Pascal would execute unless you turn off the auto-initialisation option.  Of course,
			it needed to change a little for PowerPC and for Carbon.
		*)
		var
			i: integer;
	begin
		{$ifc not TARGET_API_MAC_CARBON}
			InitGraf(@qd.thePort);
			InitFonts;
			InitWindows;
			InitMenus;
			TEInit;
			InitDialogs(nil);
			{$IFC not TARGET_RT_MAC_CFM}
				(* Guarantee that we have at least 32K of stack on 68K machines. *)
				SetApplLimit(Ptr(Get68KStackPointerRegister - 32768));
			{$ENDC}
			MaxApplZone;
		{$endc}
		for i := 1 to 3 do begin
			MoreMasters;
		end; (* for *)
	end; (* Initialization *)
	
	procedure SetupMenuBar;
		var
			mainMenuBar: Handle;
			{$ifc icDebug}
				debugMenu : MenuHandle;
			{$endc}
	begin
		mainMenuBar := GetNewMBar(rMainMenuBar);
		if mainMenuBar = nil then begin
			ExitToShell;		(* We're still not prepared to handle errors yet. *)
		end; (* if *)
		SetMenuBar(mainMenuBar);
		AppendResMenu(GetMenuHandle(mApple), 'DRVR');
		{$ifc icDebug}
			debugMenu := GetMenu(mDebug);
			ICAssert(debugMenu <> nil);
			InsertMenu(debugMenu, 0);
		{$endc}
		DrawMenuBar;
	end; (* SetupMenuBar *)
	
	procedure DieYoung(alertID : integer);
		var
			junk : integer;
	begin
		// Note that we don't use a filter here because it's probably
		// not initialised.  Besides, we're dying because we don't have
		// our minimum system requirements, which means we should minimise
		// the amount of code we execute, lest we accidentally use
		// something that isn't installed.
		
		InitCursor;
		junk := StopAlert(alertID, nil);
		ExitToShell;
	end; (* DieYoung *)

	var
		err : OSStatus;
		junk : OSStatus;
		gestaltResponse : longint;
		inst : ICInstance;
begin
	inst := nil;
	Initialization;
	
	(* Some clown is trying to run us on a system prior to System 7.  We're
		not prepare to display an error message via the normal means this early
		in the init sequence.  So instead, we use a special case error dialog.
	*)
	if not TrapAvailable(_Gestalt) | 
			(Gestalt(gestaltSystemVersion, gestaltResponse) <> noErr) | 
			(gestaltResponse < $0700) then begin
		DieYoung(rNeedSystem7Alert);
	end; (* if *)
	
	(*	Check for the Component Manager, and bail now, without trying
		anything else.
	*)

	// Carbon -- Temporarily remove this check to get it working on Carbon X.
	// Temporarily put it back until I can test this to see whether they've fixed it yet.
	
	if (Gestalt(gestaltComponentMgr, gestaltResponse) <> noErr) | (gestaltResponse = 0) then begin
		DieYoung(rNeedSystemComponentManagerAlert);
	end; (* if *)
	
	// RegisterAppearanceClient causes all our windows to be Appearance savvy,
	// even when system wide platinum is turned off on Mac OS 8.0 and Mac OS 8.1.
	// As we worked just fine without calling it (the odd cosmetic problem) on
	// those systems, we can ignore its error result.  It has no effect on Mac OS 8.5
	// and higher (except to cause us to get Appearance AppleEvents, which we will
	// have to handle eventually).
	
	if (Gestalt(gestaltAppearanceAttr, gestaltResponse) = noErr) &
							btst(gestaltResponse, gestaltAppearanceExists) then begin
		junk := RegisterAppearanceClient;
		ICAssert(junk = noErr);
	end; (* if *)
	
	(* Set up the menu bar. *)
	SetupMenuBar;
	
	(* Initialise some modules that the rest of the code depends on. *)
	InitICText;
	InitICGlobals;
	InitICMiscSubs;
	InitICIconSuites;
	InitICDialogs;
	InitICStandardFile;
	
	(* By this stage, we're prepared to Handle failure. *)
	err := InitICWindowGlobals;
	if err = noErr then begin
		err := InitICWindows;
	end; (* if *)
	if err = noErr then begin
		(* Install or update the component if necessary. *)
		(* This opens the primary connection to IC. *)
		err := InitializeComponentInstallation(inst);
	end; (* if *)
	if err = noErr then begin
		err := InitCache(inst);
	end; (* if *)
	if err = noErr then begin
		InstallAppleEventHandlers;
		err := InitICDocument(inst);
	end; (* if *)
	if err <> noErr then begin
		(* Handle an error during startup by displaying and then quitting. *)
		DisplayError(acStartApplication, err);
		ExitToShell;
	end; (* if *)
	
	InitCursor;
	while not gQuitNow do begin
		GetAndHandleAnEvent;
	end; (* while *)
	
	TermCache;
	if inst <> nil then begin
		junk := ICStop(inst);
		ICAssert(junk = noErr);
	end; (* if *)
	TermICStandardFile;				(* This calls NavUnload if necessary. *)
end. (* InternetConfig *)