unit ICComponent;

(* Start Header *)

(* File:         ICComponent.p
 * Generated by: 1.0d8
 * For:          IC 2.1
 * On:           Monday, 4 October 1999, 0:53:14
 * 
 * This file is part of the Internet Configuration system and
 * is placed in the public domain for the benefit of all.
 *)

(* End Header *)

interface

	uses
		Components,
		CodeFragments;

	const
		uppMainProcInfo = $03F0;
		
{$main}
	function Main(var params : ComponentParameters; storage : Ptr) : ComponentResult;

{$IFC TARGET_RT_MAC_CFM}

	// I have to define our own version of the RoutineDescriptor data
	// type because the CodeWarrior Pro 2 Pascal compiler wont let me
	// cast Main to a ProcPtr in order to assign it to the procDescriptor
	// field of the MainRD constant.  There were other ways to work around
	// this (for example, my frag init routine could patch up procDescriptor)
	// but this wasnt super-ugly so I decided to go this way.
	// -- Quinn, IC 2.1d2
	
	type
		MyMainEntryPoint = function(var params : ComponentParameters; storage : Ptr) : ComponentResult;
		MyRoutineRecord = RECORD
			procInfo:				ProcInfoType;							{  calling conventions  }
			reserved1:				SInt8;									{  Must be 0  }
			ISA:					ISAType;								{  Instruction Set Architecture  }
			routineFlags:			RoutineFlagsType;						{  Flags for each routine  }
			procDescriptor:			MyMainEntryPoint;						{  Where is the thing were calling?  }
			reserved2:				UInt32;									{  Must be 0  }
			selector:				UInt32;									{  For dispatched routines, the selector  }
		END;
		MyRoutineDescriptor = PACKED RECORD
			goMixedModeTrap:		UInt16;									{  Our A-Trap  }
			version:				SInt8;									{  Current Routine Descriptor version  }
			routineDescriptorFlags:	RDFlagsType;							{  Routine Descriptor Flags  }
			reserved1:				UInt32;									{  Unused, must be zero  }
			reserved2:				UInt8;									{  Unused, must be zero  }
			selectorInfo:			UInt8;									{  If a dispatched routine, calling convention, else 0  }
			routineCount:			UInt16;									{  Number of routines in this RD  }
			routineRecords:			ARRAY [0..0] OF MyRoutineRecord;		{  The individual routines  }
		END;

	// The main entry point for PowerPC code is actually a routine descriptor.
	
	const
		MainRD : MyRoutineDescriptor =
		(																				
			goMixedModeTrap:_MixedModeMagic,			// Mixed Mode A-Trap  		
			version:kRoutineDescriptorVersion,			// version 					
			routineDescriptorFlags:kSelectorsAreNotIndexable,	// RD Flags - not dispatched 	
			reserved1:0,											// reserved 1 				
			reserved2:0,											// reserved 2 				
			selectorInfo:0,											// selector info 				
			routineCount:0,											// number of routines 		
			routineRecords:(										// Its an array 				
				(										// Its a struct 				
					procInfo:uppMainProcInfo,				// the ProcInfo 				
					reserved1:0,							// reserved 					
					ISA:GetCurrentArchitecture, 			// ISA and RTA 				
					routineFlags:							// Flags
								kProcDescriptorIsAbsolute		// Its absolute addr 
								+ kFragmentIsPrepared			// Its prepared 		
								+ kUseNativeISA,				// Always use native ISA 		
					procDescriptor:Main,					// the procedure 				
					reserved2:0,							// reserved 					
					selector:0								// Not dispatched 			
				)																		
			)																		
		);

	function  ComponentFragInitFunction(var initBlock : CFragInitBlock) : OSErr;
	procedure ComponentFragTermFunction;
	
	procedure ICDummyExport;

	// Export ICDummyExport to CFM via the linker.
	
	{$LIBEXPORT+}
	{$EXPORT ICDummyExport}
	{$LIBEXPORT-}
	
{$ENDC}

implementation

	uses
		Types,
		Errors,
		Files,
		Memory, 
		Errors, 
		ToolUtils, 
		AppleTalk, 
		Aliases, 
		Packages, 
		GestaltEqu, 
		Resources, 
		Dialogs,
		AppleEvents, 
		Folders, 
		Traps,

		{$ifc TARGET_CPU_68K & not TARGET_RT_MAC_CFM & MWERKS}
			PascalA4,
		{$endc}

		InternetConfig,
		
		CFMLateImport,

		ICDebug,
		ICInstrumentation,
		ICCommonSubs,
		ICExtensionResources,
		ICExtensionGlobals,
		ICCore,
		ICURLHandling,
		ICMapping,
		ICProfiles,
		ICMultiUser;

	(* ***** Components.p Patches ***** *)

	// The following are in "Components.h" and should be in "Components.p"
	// but arent because of the way Apple auto-generates the header files.
	// So I just did a quick C to P conversion and included them here.
	
	(* MixedMode ProcInfo constants for component calls *)
	
	const
		uppComponentFunctionImplementedProcInfo			= $000002F0;
		uppGetComponentVersionProcInfo					= $000000F0;
		uppComponentSetTargetProcInfo					= $000003F0;
		uppCallComponentOpenProcInfo					= $000003F0;
		uppCallComponentCloseProcInfo					= $000003F0;
		uppCallComponentCanDoProcInfo					= $000002F0;
		uppCallComponentVersionProcInfo					= $000000F0;
		uppCallComponentRegisterProcInfo				= $000000F0;
		uppCallComponentTargetProcInfo					= $000003F0;
		uppCallComponentUnregisterProcInfo				= $000000F0;
		uppCallComponentGetMPWorkFunctionProcInfo		= $00000FF0;
		uppCallComponentGetPublicResourceProcInfo		= $00003BF0;

	(* ***** Component Manager Routines ***** *)
	
	(* Routines that are generally implemented by components. *)

	function ICCICanDo (globals: ICInstanceGlobalsPtr; selector: integer): ComponentResult;
		(* Handle the Component Manager CanDo request. We can do a standard set
			of Component Manager defined selectors, and all the selectors defined
			in ICComponentSelectors.p.
		*)
	begin
		{$unused globals}
		case selector of
			kComponentVersionSelect,
				kComponentCanDoSelect,
					kComponentOpenSelect,
					kComponentCloseSelect,
					kComponentRegisterSelect,
					kComponentUnregisterSelect,
					kComponentTargetSelect,
					kICCFirstSelector..kICCLastSelector: 
				ICCICanDo := 1;
			otherwise
				ICCICanDo := 0;
		end; (* case *)
	end; (* ICCICanDo *)

	{$ifc icInstrumentation}
		const
			gOpenCount : longint = 0;
			gInstrumentationActive : Boolean = false;
	{$endc}
	
	function ICCIOpen (globals: ICInstanceGlobalsPtr; self: ComponentInstance): ComponentResult;
		(*	Handle the Component Manager Open request.  We do very little here, basically
			using it as an opportunity to create our globals.  Most of the real initialisation
			is done in the Start routine.
			
			The globals are created in the current heap zone, not in the system heap.  This
			is very important, because, when a heap zone is destroyed (typically when an
			application quits), the Component Manager automagically cleans any instances of
			any components whose storage parameter points into that heap.  By creating our
			globals in the current heap, we ensure that a) we get cleaned up by Component
			Manager if were opened by an application, and b) if an extension wants to talk
			to IC, it can do so simply by doing a SetZone(SystemZone()) before calling
			ICStart/ICCStart.
		*)
		var
			err: ComponentResult;
	begin
		(* Create our globals. *)
		globals := ICInstanceGlobalsPtr(NewPtrClear(sizeof(ICInstanceGlobals)));
		err := MemError;

		(* Initialise some specified fields and tell the Component Manager about our globals. *)
		if err = noErr then begin
			globals^.self := self;
			globals^.target := ICInstance(self);
			SetComponentInstanceStorage(self, Handle(globals));

			if CountComponentInstances(Component(self)) = 1 then begin
				err := InitICMultiUser;
			end; (* if *)
		end; (* if *)

		{$ifc icInstrumentation}
			ICAssert(gOpenCount >= 0);
			gOpenCount := gOpenCount + 1;
			if gOpenCount = 1 then begin
				DebugStr('Starting InstrumentationLib');
				if InstInitialize68K = noErr then begin
					gInstrumentationActive := true;
					InitICMappingInstrumentation;
				end; (* if *)
			end; (* if *)
		{$endc}
		
		ICCIOpen := err;
	end; (* ICCIOpen *)

	function ICCIClose (globals: ICInstanceGlobalsPtr; self: ComponentInstance): ComponentResult;
		(*	Handle the Component Manager Close request.  We presume that the client
			has correctly called Stop, so all we have to do is dispose of our globals
			and leave.  We also Stop the instance if it hasnt been already stopped by
			the client.
		*)
		var
			junk : ICError;
	begin
		if CountComponentInstances(Component(self)) = 1 then begin
			TermICMultiUser;
		end; (* if *)
		if globals <> nil then begin
			if globals^.haveStarted then begin
				ICWatchPoint(kICDebugSignificantEventsMask, 'Close is forcing a Stop', 0);
				junk := ICCIStop(globals);
				ICAssert(junk = noErr);
				globals^.haveStarted := false;
			end; (* if *)
			DisposePtr(Ptr(globals));
			ICAssert(MemError() = noErr);
		end; (* if *)

		{$ifc icInstrumentation}
			gOpenCount := gOpenCount - 1;
			if gOpenCount = 0 then begin
				if gInstrumentationActive then begin
					DebugStr('Stopping InstrumentationLib');
					junk := InstTerminate68K;
					ICAssert( junk = noErr );
					gInstrumentationActive := false;
				end; (* if *)
			end; (* if *)
			ICAssert(gOpenCount >= 0);
		{$endc}

		ICCIClose := noErr;
	end; (* ICCIClose *)

	(*	FragRegisterFileLibs is a CFM SPI (System Programming Interface)
		that is meant to be private to Apple.  If unregister is false, it
		registers the file specified by fss as a searchable file, ie
		CFM will search it when its looking for libraries.  See the
		comment in RegisterComponentAsSharedLibraryContainer (below)
		for a description of why this is necessary.
		
		I feel this is OK to call even though its a private routine
		because, working at DTS, Ive learnt how many developers are
		using it (either because DTS told them about it, or because they
		worked it out for themselves).
	*)

	{$ifc TARGET_API_MAC_CARBON}
	
		(* 	Under Carbon, the extension file is a shared library already,
			so theres no need to register it.
		*)
		
		function FragRegisterFileLibs(var fss : FSSpec; unregister : Boolean) : OSErr;
		begin
			{$unused fss}
			{$unused unregister}
			FragRegisterFileLibs := noErr;
		end; (* FragRegisterFileLibs *)
		
	{$elsec}
		function FragRegisterFileLibs(var fss : FSSpec; unregister : Boolean) : OSErr;
			{$ifc not GENERATINGCFM}
			inline
				$3f3c, $ffec, 		// move.w #$ffec,-(sp)
				$aa5a;				// _CodeFragmentDispatch
			{$elsec}
			external;
			{$endc}
	{$endc}
	
	function RegisterOrUnregisterComponentAsSharedLibraryContainer(
									globals: ICInstanceGlobalsPtr;
									unregister : Boolean;
									var didIt : Boolean) : OSStatus;
		(*	This routine calls CFM to register the components data fork as a CFM
			shared library.  This way, our CFM based clients will be able to find
			the IC libraries in our data fork.  [Normally CFM only searches files
			of type 'shlb', but IC must necessarily be of type 'thng', so we have
			to do this trickery.
			
			Assumes that the IC resources are first in the chain.
		*)
		var
			err : OSStatus;
			extensionFileSpec : FSSpec;
	begin
		{$unused globals}
		err := noErr;
		if TrapAvailable(_CodeFragmentDispatch) then begin
			err := GetFSSpecGivenFileRefNum(CurResFile, extensionFileSpec);
			if err = noErr then begin
				err := FragRegisterFileLibs(extensionFileSpec, unregister);
				didIt := (err = noErr);
				if not unregister & (err = cfragDupRegistrationErr) then begin
					ICWatchPoint(kICDebugSignificantEventsMask, 'Swallowing duplicate register error', 0);
					err := noErr;
				end; (* if *)
			end; (* if *)
		end; (* if *)
		RegisterOrUnregisterComponentAsSharedLibraryContainer := err;
	end; (* RegisterOrUnregisterComponentAsSharedLibraryContainer *)

	{$ifc TARGET_RT_MAC_CFM}
		var
			gFragFile         : FSSpec;
			gFragLocator      : CFragSystem7DiskFlatLocator;
			gFragConnID  	  : CFragConnectionID;
			gLateImportConnID : CFragConnectionID;

		function LateImportInternetConfigLib(var connID : CFragConnectionID) : OSStatus;
			var
				err : OSStatus;
				junk : OSStatus;
				oldZone : THz;
				junkMain : Ptr;
		begin
			ICAssert(@ICGetPref = Ptr(kUnresolvedCFragSymbolAddress));
			
			connID := nil;
			
			oldZone := GetZone;
			SetZone(SystemZone);
			
			// Open up a connection to 'InternetConfigLib' and late bind to it.
			
			err := GetSharedLibrary('ICAp;InternetConfigLib', kCompiledCFragArch, kReferenceCFrag,
										connID, junkMain, nil);
			if err = noErr then begin
				err := CFMLateImportLibrary(@gFragLocator, gFragConnID,
												'ICDummyExport',
												'ICAp;InternetConfigLib',
												connID);
			end; (* if *)
			
			// Clean up.
			
			if (err <> noErr) & (connID <> nil) then begin
				junk := CloseConnection(connID);
				ICAssert(junk = noErr);
			end; (* if *)
			SetZone(oldZone);
			
			ICAssert(((err = noErr) & (connID <> nil)) | ((err <> noErr) & (connID = nil)));
			
			LateImportInternetConfigLib := err;
		end; (* LateImportInternetConfigLib *)

	{$endc}

	function ICCIRegister(globals : ICInstanceGlobalsPtr) : ComponentResult;
		var
			err : ComponentResult;
			err2 : ComponentResult;
			junk : OSStatus;
			gestaltValue : longint;
			didOpen : Boolean;
			haveRegistered : Boolean;
			junkBool : Boolean;
	begin
		err := noErr;
		if not TrapAvailable(_Gestalt) | 
				(Gestalt(gestaltSystemVersion, gestaltValue) <> noErr) | 
				(gestaltValue < $0700) then begin
			err := unimpErr;
		end; (* if *)
		if err = noErr then begin

			// Now create or update the system-wide globals (which are
			// accessed via Gestalt).  The sequencing of this is somewhat
			// tricky.  We open up the component resource file early
			// because a) we need it open to call RegisterOrUnregisterComponentAsSharedLibraryContainer,
			// and b) we want it open while we call RegisterOrUpdateSystemGlobals
			// because RegisterOrUpdateSystemGlobals may need it open.
			// [All code opens the resources with StartUsingICResources,
			// which reference counts, so having RegisterOrUpdateSystemGlobals
			// open it a second time is both safe and cheap.]
			//
			// Then we register the components data fork as a CFM container.
			// We need to do this the first time around because the container
			// wont be registered.  We need to do this the second time around,
			// even if the system globals already exist, mostly to support
			// hot-replace (especially useful while debugging).
			//
			// After registering, we update the system globals to our version.
			//
			// Then we fix up the CFM imports of IC.  The components fragment is
			// weak linked to "InternetConfigLib" so that it will load the first
			// time around, before weve registered the data fork as a CFM container.
			// If the weak link failed, we want to fix it up because the rest
			// of the IC code expects to be able to call IC routines without testing.
			
			haveRegistered := false;
			err := StartUsingICResources(globals, didOpen);
			if err = noErr then begin
				err := RegisterOrUnregisterComponentAsSharedLibraryContainer(globals, false, haveRegistered);
				if (err = noErr) & (FindICSystemGlobals = nil) then begin
					if err = noErr then begin
						err := RegisterOrUpdateSystemGlobals(globals);
					end; (* if *)
				end; (* if *)
				if err = noErr then begin
					{$ifc TARGET_RT_MAC_CFM & not TARGET_API_MAC_CARBON}
						if @ICGetPref = Ptr(kUnresolvedCFragSymbolAddress) then begin
							ICWatchPoint(kICDebugSignificantEventsMask, 'Register requires a late bind', 0);
							ICAssert(gLateImportConnID = nil);
							err := LateImportInternetConfigLib(gLateImportConnID);
						end; (* if *)
					{$endc}
				end; (* if *)
			end; (* if *)

			// Clean up.
			
			err2 := StopUsingICResources(globals, didOpen);
			if err = noErr then begin
				err := err2;
			end; (* if *)

			// Clean up CFM registration on error.

			if err <> noErr then begin
				if haveRegistered then begin
					junk := RegisterOrUnregisterComponentAsSharedLibraryContainer(globals, true, junkBool);
					ICAssert((junk = noErr) & junkBool);
				end; (* if *)				
			end; (* if *)
		end; (* if *)
		if err <> noErr then begin
			err := 1;					// Inside Mac says the Register routine should return true to not be registered
		end; (* if *)
		ICCIRegister := err;
	end; (* ICCIRegister *)

	function ICCIUnregister(globals : ICInstanceGlobalsPtr) : ComponentResult;
		var
			err : OSStatus;
			err2 : OSStatus;
			didOpen : Boolean;
			junkBool : Boolean;
	begin
		// First unregister our data fork as a CFM container.
		
		err := StartUsingICResources(globals, didOpen);
		if err = noErr then begin
			err := RegisterOrUnregisterComponentAsSharedLibraryContainer(globals, true, junkBool);
		end; (* if *)
		err2 := StopUsingICResources(globals, didOpen);
		if err = noErr then begin
			err := err2;
		end; (* if *)

		// Failure to unregister the extension with CFM is not really a fatal
		// error.  This code only really exists to help me with debugging.  In
		// a production system, the component very rarely gets unregistered
		// (only when the IC application hot installs a new version).

		ICAssert(err = noErr);
		
		ICCIUnregister := noErr;
	end; (* ICCIUnregister *)

	function ICCITarget (globals: ICInstanceGlobalsPtr; newTarget: ComponentInstance): ComponentResult;
		(* Handle the Component Manager Target.  We change our current target
			variable to reflect the new target.
		*)
	begin
		globals^.target := ICInstance(newTarget);
		ICCITarget := noErr;
	end; (* ICCITarget *)

(* ----- Component Entry Point ----- *)

	function Main(var params : ComponentParameters; storage : Ptr) : ComponentResult;
		(* The main component entry point.  Its just a standard component dispatcher. *)

		{$ifc icDebug}

			procedure ComponentEntryExitDebug(entry : Boolean; result : ComponentResult);
				var
					sysGlobals : ICSystemGlobalsPtr;
			begin
				if storage <> nil then begin
					sysGlobals := FindICSystemGlobals;
					if sysGlobals <> nil then begin
						if entry then begin
							if band(sysGlobals^.debugFlags, kICDebugComponentEntryMask) <> 0 then begin
								DebugStr("IC Component Entry " + SelectorToStr(params.what));
							end; (* if *)
						end else begin
							if band(sysGlobals^.debugFlags, kICDebugComponentExitMask) <> 0 then begin
								DebugStr("IC Component Exit  " + SelectorToStr(params.what) + " " + DecStr(result));
							end; (* if *)
						end; (* if *)
					end; (* if *)
				end; (* if *)
			end; (* ComponentEntryExitDebug *)

		{$endc}

		var
			{$ifc TARGET_CPU_68K & not TARGET_RT_MAC_CFM & MWERKS}
				oldA4 : longint;
			{$endc}
			oldA5 : longint;
			proc: ProcPtr;
			procInfo : ProcInfoType; 
			result : ComponentResult;
	begin
		(*	If were being compiled under Metrowerks, bring in the correct A4 world.
			Under both Metrowerks and MPW, set up A5 to point to the current applications
			A5 world.  IC doesnt use A5 based globals, except insofar as system API
			routines use them.  So we dont need to create our own A5 world.  But we
			do need to make sure our A5 world is correct, because we call system
			routines (like ModalDialog, which calls QuickDraw, and Unique1ID, which
			calls Random) that need A5 to point to some valid A5 world.  A5 may not
			be pointing to a valid A5 world in all possible in which were called
			(the case that triggered this change was IC running under Mac OS 8 with
			the NetLaunchURL override component installed, where A5 was nil) so
			we make sure its the case here.
		*)
		{$ifc TARGET_CPU_68K & not TARGET_RT_MAC_CFM & MWERKS} oldA4 := SetCurrentA4; {$endc}
		oldA5 := SetCurrentA5;

		{$ifc icDebug}
			ComponentEntryExitDebug(true, 0);
		{$endc}
		
		(* Setup proc based on which routine is to be called. *)
		
		proc := nil;
		procInfo := 0;
		case params.what of
			(* Component Manager stuff *)
			kComponentVersionSelect: 
				result := kICCurrentComponentVersion;
			kComponentCanDoSelect: 
				begin proc := @ICCICanDo; procInfo := uppCallComponentCanDoProcInfo; end;
			kComponentOpenSelect: 
				begin proc := @ICCIOpen; procInfo := uppCallComponentOpenProcInfo; end;
			(* Component Manager stuff *)
			kComponentCloseSelect: 
				begin proc := @ICCIClose; procInfo := uppCallComponentCloseProcInfo; end;
			kComponentRegisterSelect:
				begin proc := @ICCIRegister; procInfo := uppCallComponentRegisterProcInfo; end;
			kComponentUnregisterSelect:
				begin proc := @ICCIUnregister; procInfo := uppCallComponentUnregisterProcInfo; end;
			kComponentTargetSelect: 
				begin proc := @ICCITarget; procInfo := uppCallComponentTargetProcInfo; end;
			(* this component type stuff *)
(* Start ICCSwitch.p *)

			kICCStart: 
			    begin proc := @ICCIStart; procInfo := kICCStartProcInfo; end;
			kICCStop: 
			    begin proc := @ICCIStop; procInfo := kICCStopProcInfo; end;
			kICCGetVersion: 
			    begin proc := @ICCIGetVersion; procInfo := kICCGetVersionProcInfo; end;
			{$ifc CALL_NOT_IN_CARBON}
			kICCFindConfigFile: 
			    begin proc := @ICCIFindConfigFile; procInfo := kICCFindConfigFileProcInfo; end;
			{$endc}
			{$ifc CALL_NOT_IN_CARBON}
			kICCFindUserConfigFile: 
			    begin proc := @ICCIFindUserConfigFile; procInfo := kICCFindUserConfigFileProcInfo; end;
			{$endc}
			{$ifc CALL_NOT_IN_CARBON}
			kICCGeneralFindConfigFile: 
			    begin proc := @ICCIGeneralFindConfigFile; procInfo := kICCGeneralFindConfigFileProcInfo; end;
			{$endc}
			{$ifc CALL_NOT_IN_CARBON}
			kICCChooseConfig: 
			    begin proc := @ICCIChooseConfig; procInfo := kICCChooseConfigProcInfo; end;
			{$endc}
			{$ifc CALL_NOT_IN_CARBON}
			kICCChooseNewConfig: 
			    begin proc := @ICCIChooseNewConfig; procInfo := kICCChooseNewConfigProcInfo; end;
			{$endc}
			kICCGetConfigName: 
			    begin proc := @ICCIGetConfigName; procInfo := kICCGetConfigNameProcInfo; end;
			{$ifc CALL_NOT_IN_CARBON}
			kICCGetConfigReference: 
			    begin proc := @ICCIGetConfigReference; procInfo := kICCGetConfigReferenceProcInfo; end;
			{$endc}
			{$ifc CALL_NOT_IN_CARBON}
			kICCSetConfigReference: 
			    begin proc := @ICCISetConfigReference; procInfo := kICCSetConfigReferenceProcInfo; end;
			{$endc}
			{$ifc CALL_NOT_IN_CARBON}
			kICCSpecifyConfigFile: 
			    begin proc := @ICCISpecifyConfigFile; procInfo := kICCSpecifyConfigFileProcInfo; end;
			{$endc}
			{$ifc CALL_NOT_IN_CARBON}
			kICCRefreshCaches: 
			    begin proc := @ICCIRefreshCaches; procInfo := kICCRefreshCachesProcInfo; end;
			{$endc}
			kICCGetSeed: 
			    begin proc := @ICCIGetSeed; procInfo := kICCGetSeedProcInfo; end;
			kICCGetPerm: 
			    begin proc := @ICCIGetPerm; procInfo := kICCGetPermProcInfo; end;
			{$ifc CALL_NOT_IN_CARBON}
			kICCDefaultFileName: 
			    begin proc := @ICCIDefaultFileName; procInfo := kICCDefaultFileNameProcInfo; end;
			{$endc}
			kICCBegin: 
			    begin proc := @ICCIBegin; procInfo := kICCBeginProcInfo; end;
			kICCGetPref: 
			    begin proc := @ICCIGetPref; procInfo := kICCGetPrefProcInfo; end;
			kICCSetPref: 
			    begin proc := @ICCISetPref; procInfo := kICCSetPrefProcInfo; end;
			kICCFindPrefHandle: 
			    begin proc := @ICCIFindPrefHandle; procInfo := kICCFindPrefHandleProcInfo; end;
			kICCGetPrefHandle: 
			    begin proc := @ICCIGetPrefHandle; procInfo := kICCGetPrefHandleProcInfo; end;
			kICCSetPrefHandle: 
			    begin proc := @ICCISetPrefHandle; procInfo := kICCSetPrefHandleProcInfo; end;
			kICCCountPref: 
			    begin proc := @ICCICountPref; procInfo := kICCCountPrefProcInfo; end;
			kICCGetIndPref: 
			    begin proc := @ICCIGetIndPref; procInfo := kICCGetIndPrefProcInfo; end;
			kICCDeletePref: 
			    begin proc := @ICCIDeletePref; procInfo := kICCDeletePrefProcInfo; end;
			kICCEnd: 
			    begin proc := @ICCIEnd; procInfo := kICCEndProcInfo; end;
			kICCGetDefaultPref: 
			    begin proc := @ICCIGetDefaultPref; procInfo := kICCGetDefaultPrefProcInfo; end;
			kICCEditPreferences: 
			    begin proc := @ICCIEditPreferences; procInfo := kICCEditPreferencesProcInfo; end;
			kICCParseURL: 
			    begin proc := @ICCIParseURL; procInfo := kICCParseURLProcInfo; end;
			kICCLaunchURL: 
			    begin proc := @ICCILaunchURL; procInfo := kICCLaunchURLProcInfo; end;
			kICCMapFilename: 
			    begin proc := @ICCIMapFilename; procInfo := kICCMapFilenameProcInfo; end;
			kICCMapTypeCreator: 
			    begin proc := @ICCIMapTypeCreator; procInfo := kICCMapTypeCreatorProcInfo; end;
			kICCMapEntriesFilename: 
			    begin proc := @ICCIMapEntriesFilename; procInfo := kICCMapEntriesFilenameProcInfo; end;
			kICCMapEntriesTypeCreator: 
			    begin proc := @ICCIMapEntriesTypeCreator; procInfo := kICCMapEntriesTypeCreatorProcInfo; end;
			kICCCountMapEntries: 
			    begin proc := @ICCICountMapEntries; procInfo := kICCCountMapEntriesProcInfo; end;
			kICCGetIndMapEntry: 
			    begin proc := @ICCIGetIndMapEntry; procInfo := kICCGetIndMapEntryProcInfo; end;
			kICCGetMapEntry: 
			    begin proc := @ICCIGetMapEntry; procInfo := kICCGetMapEntryProcInfo; end;
			kICCSetMapEntry: 
			    begin proc := @ICCISetMapEntry; procInfo := kICCSetMapEntryProcInfo; end;
			kICCDeleteMapEntry: 
			    begin proc := @ICCIDeleteMapEntry; procInfo := kICCDeleteMapEntryProcInfo; end;
			kICCAddMapEntry: 
			    begin proc := @ICCIAddMapEntry; procInfo := kICCAddMapEntryProcInfo; end;
			kICCGetCurrentProfile: 
			    begin proc := @ICCIGetCurrentProfile; procInfo := kICCGetCurrentProfileProcInfo; end;
			kICCSetCurrentProfile: 
			    begin proc := @ICCISetCurrentProfile; procInfo := kICCSetCurrentProfileProcInfo; end;
			kICCCountProfiles: 
			    begin proc := @ICCICountProfiles; procInfo := kICCCountProfilesProcInfo; end;
			kICCGetIndProfile: 
			    begin proc := @ICCIGetIndProfile; procInfo := kICCGetIndProfileProcInfo; end;
			kICCGetProfileName: 
			    begin proc := @ICCIGetProfileName; procInfo := kICCGetProfileNameProcInfo; end;
			kICCSetProfileName: 
			    begin proc := @ICCISetProfileName; procInfo := kICCSetProfileNameProcInfo; end;
			kICCAddProfile: 
			    begin proc := @ICCIAddProfile; procInfo := kICCAddProfileProcInfo; end;
			kICCDeleteProfile: 
			    begin proc := @ICCIDeleteProfile; procInfo := kICCDeleteProfileProcInfo; end;
			kICCRequiresInterruptSafe: 
			    begin proc := @ICCIRequiresInterruptSafe; procInfo := kICCRequiresInterruptSafeProcInfo; end;
			kICCGetMappingInterruptSafe: 
			    begin proc := @ICCIGetMappingInterruptSafe; procInfo := kICCGetMappingInterruptSafeProcInfo; end;
			kICCGetSeedInterruptSafe: 
			    begin proc := @ICCIGetSeedInterruptSafe; procInfo := kICCGetSeedInterruptSafeProcInfo; end;

(* End ICCSwitch.p *)
			otherwise
				result := badComponentSelector;
		end; (* case *)

		(* If weve got a routine to call, call it.  Note that the default case
			above has set result up to return the appropriate error code if
			proc is nil.
		*)
		if proc <> nil then begin
			ICAssert(procInfo <> 0);
			{$ifc TARGET_RT_MAC_CFM}
				result := CallComponentFunctionWithStorageProcInfo(Handle(storage), params, proc, procInfo);
			{$elsec}
				result := CallComponentFunctionWithStorage(Handle(storage), params, proc);
			{$endc}
		end; (* if *)

		(* Clean up A5 and A4. *)

		{$ifc icDebug}
			ComponentEntryExitDebug(false, result);
		{$endc}
		
		oldA5 := SetA5(oldA5);
		{$ifc TARGET_CPU_68K & not TARGET_RT_MAC_CFM & MWERKS} oldA4 := SetA4(oldA4); {$endc}
		Main := result;
	end; (* Main *)

	{$ifc TARGET_RT_MAC_CFM}

		procedure ICDummyExport;
		begin
			ICDebugStr('ICDummyExport: Who is calling us?');
		end; (* ICDummyExport *)

		function  ComponentFragInitFunction(var initBlock : CFragInitBlock) : OSErr;
			var
				err : OSStatus;
		begin
			if initBlock.fragLocator.where = kDataForkCFragLocator then begin
				gFragLocator := initBlock.fragLocator.onDisk;
				gFragFile := initBlock.fragLocator.onDisk.fileSpec^;
				gFragLocator.fileSpec := @gFragFile;
				gFragConnID := initBlock.connectionID;
				gLateImportConnID := nil;
				err := noErr;
			end else begin
				ICDebugStr('ComponentFragInitFunction: We are not being loaded from a data fork!');
				err := -1;
			end; (* if *)
			
			ComponentFragInitFunction := err;
		end; (* ComponentFragInitFunction *)

		procedure ComponentFragTermFunction;
			var
				junk : OSStatus;
				oldZone : THz;
		begin
			if gLateImportConnID <> nil then begin
				oldZone := GetZone;
				SetZone(SystemZone);
				junk := CloseConnection(gLateImportConnID);
				ICAssert(junk = noErr);
				ICAssert(gLateImportConnID = nil);		// CloseConnection is supposed to clear it for us.
				SetZone(oldZone);
			end; (* if *)
		end; (* ComponentFragTermFunction *)

	{$endc}
	
end. (* ICComponent *)