unit ICMapping;

(* Start Header *)

(* File:         ICMapping.p
 * Generated by: 1.0d8
 * For:          IC 2.1
 * On:           Monday, 4 October 1999, 0:53:14
 * 
 * This file is part of the Internet Configuration system and
 * is placed in the public domain for the benefit of all.
 *)

(* End Header *)

(* This file contains the core implementation of the Internet Config API.
	It defines the format of the ICRRecord type (which the outside world
	sees indirectly by way of the ICInstance, a pointer to this record) and
	the various IC API operations on that type.
*)

interface

	uses
		Types,
		Files, 
		QuickDraw, 
		Aliases, 
		Memory,
		Components,
		
		InternetConfig,
		ICInstrumentation,
		ICCore;

	(* ***** Mapping API Routines ***** *)
	
	(*	These interfaces are automagically generated by the Internet Config
		Interface Generator.  See "InternetConfig.p" for a description of
		the behaviour of these routines.
	*)

(* Start ICCIMapping.p *)

	function ICCIMapFilename(globals : ICInstanceGlobalsPtr; filename : StringPtr; var entry : ICMapEntry) : ICError;
	
	function ICCIMapTypeCreator(globals : ICInstanceGlobalsPtr; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
	
	function ICCIMapEntriesFilename(globals : ICInstanceGlobalsPtr; entries : Handle; filename : StringPtr; var entry : ICMapEntry) : ICError;
	
	function ICCIMapEntriesTypeCreator(globals : ICInstanceGlobalsPtr; entries : Handle; fType : OSType; fCreator : OSType; filename : StringPtr; var entry : ICMapEntry) : ICError;
	
	function ICCICountMapEntries(globals : ICInstanceGlobalsPtr; entries : Handle; var count : longint) : ICError;
	
	function ICCIGetIndMapEntry(globals : ICInstanceGlobalsPtr; entries : Handle; index : longint; var pos : longint; var entry : ICMapEntry) : ICError;
	
	function ICCIGetMapEntry(globals : ICInstanceGlobalsPtr; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
	
	function ICCISetMapEntry(globals : ICInstanceGlobalsPtr; entries : Handle; pos : longint; var entry : ICMapEntry) : ICError;
	
	function ICCIDeleteMapEntry(globals : ICInstanceGlobalsPtr; entries : Handle; pos : longint) : ICError;
	
	function ICCIAddMapEntry(globals : ICInstanceGlobalsPtr; entries : Handle; var entry : ICMapEntry) : ICError;
	
	function ICCIRequiresInterruptSafe(globals : ICInstanceGlobalsPtr) : ICError;
	
	function ICCIGetMappingInterruptSafe(globals : ICInstanceGlobalsPtr; var mappingPref : Ptr; var mappingPrefSize : longint) : ICError;
	
	function ICCIGetSeedInterruptSafe(globals : ICInstanceGlobalsPtr; var seed : longint) : ICError;
	

(* End ICCIMapping.p *)

	function SetInterrruptSafeMapping(sysGlobals : ICSystemGlobalsPtr; entries : Ptr; entriesSize : longint) : ICError;
	function RebuildInterruptSafeMappingSysGlobals(sysGlobals : ICSystemGlobalsPtr; inst : ICInstance) : ICError;
	function RebuildInterruptSafeMapping(globals : ICInstanceGlobalsPtr) : ICError;

	{$ifc icInstrumentation}
		procedure InitICMappingInstrumentation;
	{$endc}

implementation

	uses
		Resources,
		GestaltEqu, 
		OSUtils, 
		ToolUtils, 
		StandardFile, 
		AppleTalk, 
		Folders, 
		Processes, 
		Errors, 
		AppleEvents, 
		LowMem,
		TextUtils,

		ICDebug,
		ICStringLists,
		ICCommonSubs,
		ICResourceForkSanity, 
		ICCore,
		ICExtensionResources;

	(* ***** Internal Mapping Routines *****)

	const
		kMinimumICMapEntryLength = 6;
		(* An ICMapEntry must have at least 6 bytes in it, namely
			the total_length, fixed_length and version fields.
		*)

	GLOBAL_TRACE_SETUP(gUnpackEntryInterruptSafe)
	GLOBAL_TRACE_SETUP(gCopyStringOut)

	function UnpackEntryInterruptSafe(entries: Ptr; entriesSize : longint; 
							pos: longint; 
							var entry: ICMapEntry; 
							userDataBuffer : Ptr; var userDataSize: longint): OSErr;
		(* This routine unpacks the ICMapEntry from position pos in the entries
			'array'.  It fills out the unpacked form of the ICMapEntry
			and the userDataSize count, which is the number of bytes
			of user data in the entry.  Also, if userDataBuffer is not
			nil, it copies out the user data to that buffer, whose
			maximum size is determined by the incoming value of userDataSize.
			userDataSize is ignored if userDataBuffer is nil.
			
			WARNING: 	Depends very much on the exact format of ICMapEntry!

			WARNING: 	This routine is called at interrupt time, and should not
						move, purge, rely on unlocked handles, make synchronous
						calls, or anything else that is not interrupt safe.
			
			NOTE:		Interrupt safe, in this context, does not mean VM
						safe.  This routine is allowed to cause page faults.
			
			IMPORTANT:	You need to read the comment "Notes About Interrupt
						Safe API Changes", later in this file.
		*)

		procedure CopyStringOut (var p: Ptr; var s: Str255; copyString : Boolean);
			(* Copies a Pascal string from p to s using BlockMoveData
				to avoid alignment problems.  It also updates p to point
				to the byte after the string.
			*)
			var
				stringLength: integer;
				TRACE_SETUP
		begin
			LOG_ENTRY(gCopyStringOut);
			stringLength := band(p^, $0FF) + 1;
			if copyString then begin
				BlockMoveData(p, @s, stringLength);
			end else begin
				s[0] := chr(0);
			end; (* if *)
			p := Ptr(ord(p) + stringLength);
			LOG_EXIT(gCopyStringOut);
		end; (* CopyStringOut *)

		var
			err: OSErr;
			entryBase: Ptr;
			copyCursor: Ptr;
			userDataBufferSize : longint;
			TRACE_SETUP
	begin
		LOG_ENTRY(gUnpackEntryInterruptSafe);
		(* Check parameters.  Because this internal routine is the first
			'point-of-contact' that many of the IC routines have to parameters
			passed in by the client, we do rigorous parameter checking here
			to save doing it in all of the IC routines that call here.
		*)
		err := noErr;
		ICParamCheck(err, entries <> nil, 'UnpackEntryInterruptSafe: entries is nil');
		ICParamCheck(err, pos >= 0, 'UnpackEntryInterruptSafe: pos is negative');
		ICParamCheck(err, pos <= (entriesSize - kMinimumICMapEntryLength), 'UnpackEntryInterruptSafe: pos is off the end of entries');

		(* Get the first 6 bytes out of the entry and do a rough check for
			their validity.  This makes sure we don't go off into the weeds
			when passed a completely bogus entries handle.
		*)
		if err = noErr then begin
			entryBase := Ptr(ord(entries) + pos);
			BlockMoveData(entryBase, @entry, kMinimumICMapEntryLength);
			if (entry.fixed_length <> ICmap_fixed_length) |
						(entry.fixed_length > entry.total_length) | 
								(entry.total_length > entriesSize) then begin
				(* WTF is this error code!  See what happens when you
					let PNL write the code!
				*)
				err := badExtResource;
			end; (* if *)
		end; (* if *)
		
		(* All is good.  Copy the data out of the entries handle
			into the ICMapEntry, unpacking the packed Pascal
			strings as we go.  Use copyCursor to keep track
			of the current position into the variable length
			data as we copy the data out.  At the end, 
			use the difference between copyCursor and entryBase
			to calculate the total bytes that we know about
			in the entry and hence the number of user bytes.
		*)
		if err = noErr then begin
			BlockMoveData(entryBase, @entry, entry.fixed_length);
			copyCursor := Ptr(ord(entryBase) + entry.fixed_length);
			CopyStringOut(copyCursor, entry.extension, true);
			CopyStringOut(copyCursor, entry.creator_app_name, true);
			CopyStringOut(copyCursor, entry.post_app_name, true);
			CopyStringOut(copyCursor, entry.MIME_type, true);
			CopyStringOut(copyCursor, entry.entry_name, true);
			userDataBufferSize := userDataSize;
			userDataSize := entry.total_length - (ord(copyCursor) - ord(entryBase));
			if userDataBuffer <> nil then begin
				if userDataBufferSize > userDataSize then begin
					userDataBufferSize := userDataSize;
				end; (* if *)
				BlockMoveData(copyCursor, userDataBuffer, userDataBufferSize);
			end; (* if *)
		end; (* if *)
		LOG_EXIT(gUnpackEntryInterruptSafe);
		UnpackEntryInterruptSafe := err;
	end; (* UnpackEntryInterruptSafe *)

	function UnpackEntry (entries: Handle; pos: longint; var entry: ICMapEntry; var userDataSize: longint): OSErr;
		var
			err : ICError;
			s : SInt8;
	begin
		s := HGetState(entries);
		HLock(entries);
		userDataSize := 0;
		err := UnpackEntryInterruptSafe(entries^, GetHandleSize(entries), pos,
					entry, nil, userDataSize);
		HSetState(entries, s);
		UnpackEntry := err;
	end; (* UnpackEntry *)

	procedure PackEntry (var entry: ICMapEntry; p: Ptr; userDataSize: longint);
		(* Pack the ICMapEntry and copy it into the address pointed to by p.
			When setting up the entry.total_length field, leave enough
			space for userDataSize bytes of user data.
		*)

		procedure CopyStringIn (var s: Str255);
			(* Copy the string s into the client buffer pointed to by
				p in a packed fashion, using BlockMoveData to avoid
				alignment problems.  Update entry.total_length
				to ensure that the next call to this routine will
				copy the string packed immediately after this string.
			*)
		begin
			BlockMoveData(@s, Ptr(ord(p) + entry.total_length), length(s) + 1);
			entry.total_length := entry.total_length + length(s) + 1;
		end; (* CopyStringIn *)

	begin
		entry.version := 0;
		entry.fixed_length := ord(@entry.extension) - ord(@entry);

		(* Before copying the strings in, set up entry.total_length
			to make sure that CopyStringIn starts copying to the right
			place.
		*)
		entry.total_length := entry.fixed_length;
		CopyStringIn(entry.extension);
		CopyStringIn(entry.creator_app_name);
		CopyStringIn(entry.post_app_name);
		CopyStringIn(entry.MIME_type);
		CopyStringIn(entry.entry_name);
		
		(* Now that we've finished copying the strings in, update
			entry.total_length to include the number of bytes of user data.
			Then copy the fixed fields of the entry into the user buffer.
		*)
		entry.total_length := entry.total_length + userDataSize;
		BlockMoveData(@entry, p, entry.fixed_length);
	end; (* PackEntry *)

	GLOBAL_TRACE_SETUP(gFastGetEntry)
	
	function FastGetEntry (entries: Handle; pos: longint; var entry: ICMapEntry): OSErr;
		(* A faster version of UnpackEntry, which only returns the fixed part
			of the entry along with the extension string.  This is called by
			clients which need only this information and need to go quickly.

			WARNING: Depends very much on the exact format of ICMapEntry!
		*)
		var
			err: OSErr;
			entryBase: Ptr;
			maxSize: longint;
			p: Ptr;
			TRACE_SETUP
	begin
		LOG_ENTRY(gFastGetEntry);
		(* Check parameters.  Because this internal routine is the first
			'point-of-contact' that many of the IC routines have to parameters
			passed in by the client, we do rigorous parameter checking here
			to save doing it in all of the IC routines that call here.
		*)
		err := noErr;
		ICParamCheck(err, entries <> nil, 'FastGetEntry: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'FastGetEntry: entries has been purged');
		ICParamCheck(err, pos >= 0, 'FastGetEntry: pos is negative');
		
		// The following two checks are mostly equivalent.  I've substituted
		// the more complex one because its common for applications (and
		// indeed IC itself) to repeatedly call this entry, incrementing pos,
		// until they get back the paramErr.  So we need to isolate the
		// terminating condition (pos = handle size) from a possible
		// application bug (pos > handle size).
		
		// ICParamCheck(err, pos < GetHandleSize(entries), 'FastGetEntry: pos is off the end of entries');
		
		if pos >= GetHandleSize(entries) then begin
			{$ifc icDebug}
				if pos > GetHandleSize(entries) then begin
					ICDebugStr('FastGetEntry: pos is off the end of entries');
				end; (* if *)
			{$endc}
			err := paramErr;
		end; (* if *)
	
		(* Get the first 6 bytes out of the entry and do a rough check for
			their validity.  This makes sure we don't go off into the weeds
			when passed a completely bogus entries handle.
		*)
		if err = noErr then begin
			entryBase := (Ptr(ord(entries^) + pos));
			maxSize := GetHandleSize(entries);
			BlockMoveData(entryBase, @entry, kMinimumICMapEntryLength);
			if (entry.fixed_length <> ICmap_fixed_length) |
						(entry.fixed_length > entry.total_length) |
								(entry.total_length > maxSize) then begin
				err := badExtResource;
			end; (* if *)
		end; (* if *)

		(* All is good.  Copy the fixed data out of the entries handle
			into the ICMapEntry.  Also copy out the extension field,
			which is really easy to find and used by the clients
			of this routine.
		*)
		if err = noErr then begin
			BlockMoveData(entryBase, @entry, entry.fixed_length);
			p := Ptr(ord(entryBase) + entry.fixed_length);
			BlockMoveData(p, @entry.extension, band(p^, $0FF) + 1);
		end; (* if *)
		LOG_EXIT(gFastGetEntry);
		FastGetEntry := err;
	end; (* FastGetEntry *)

	function GetShort (p: Ptr): integer;
		(* Gets a short integer out of an arbitrary location,
			without worrying about alignment restrictions.
			This would be a good candidate for inlining (-:
		*)
	begin
		GetShort := band(p^, $0FF) * 256 + band(Ptr(ord(p) + 1)^, $0FF);
	end; (* if *)

	{$ifc TARGET_RT_MAC_CFM}
		function UpCase(ch : char) : char;
		begin
			if ch in ['a'..'z'] then begin
				UpCase := chr(ord(ch) - (ord('a') - ord('A')));
			end else begin
				UpCase := ch;
			end; (* if *)
		end; (* UpCase *)
	{$elsec}
		function UpCase (ch: char): char;
			(* Upper case an ASCII character the quick and dirty way. *)
			inline
				$301F,					(* 				move.w		(sp)+,d0 						*)
				$0C00, $0061,			(* 				cmpi.b		#'a',d0							*)
				$6500, $000E,			(* 				bcs			done				; d0 < 'a'	*)
				$0C00, $007B,			(* 				cmpi.b		#'{',d0							*)
				$6400, $0006,			(* 				bcc			done				; d0 >= '{'	*)
				$0400, $0020,			(* 				subi.b		#$20,d0							*)
				$3E80;					(* done:		move.w		d0,(a7)							*)
	{$endc}

	function IsExtensionVar (var name, extension: Str255): Boolean;
		(* Returns true if name has a file extension that matches extension. *)
		var
			nameIndex : integer;
			extensionIndex: integer;
	begin
		IsExtensionVar := false;
		if length(name) >= length(extension) then begin
			nameIndex := length(name) - length(extension) + 1;
			extensionIndex := 1;
			while extensionIndex <= length(extension) do begin
				if UpCase(name[nameIndex]) <> UpCase(extension[extensionIndex]) then begin
					leave;
				end; (* if *)
				nameIndex := nameIndex + 1;
				extensionIndex := extensionIndex + 1;
			end; (* while *)
			IsExtensionVar := (extensionIndex > length(extension));
		end; (* if *)
	end; (* IsExtensionVar *)

	(* ***** Low Level Mapping Routines ***** *)

	function ICCICountMapEntries (globals : ICInstanceGlobalsPtr; entries: Handle; var count: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entryBase: Ptr;
			pos: longint;
			entriesHandleSize : longint;
			entrySize: integer;
	begin
		(* Check parameters. *)
		{$unused globals}
		err := noErr;
		ICParamCheck(err, entries <> nil, 'ICCountMapEntries: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'ICCountMapEntries: entries has been purged');
		
		(* Starting at the beginning of the entries handle, extract the total_length
			field to skip to the next entry, counting the entries as we go, until
			we run out of entries in the handle.
		*)
		if err = noErr then begin
			entryBase := entries^;
			pos := 0;
			count := 0;
			entriesHandleSize := GetHandleSize(entries);
			while pos < entriesHandleSize do begin
				entrySize := GetShort(entryBase);
				pos := pos + entrySize;
				entryBase := Ptr(ord(entryBase) + entrySize);
				count := count + 1;
			end; (* while *)
		end; (* if *)
		ICCICountMapEntries := err;
	end; (* ICCICountMapEntries *)

	function ICCIGetIndMapEntry (globals : ICInstanceGlobalsPtr; entries: Handle; ndx: longint; var pos: longint; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entryBase: Ptr;
			entrySize: integer;
			entriesHandleSize : longint;
	begin
		(* Check parameters. *)
		err := noErr;
		ICParamCheck(err, entries <> nil, 'ICGetIndMapEntry: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'ICGetIndMapEntry: entries has been purged');
		
		(* Skip through the handle (in a manner similar to that
			used by CountMapEntries) looking for the ndx'th
			entry.  When we find it, call GetMapEntry to extract it.
		*)
		if err = noErr then begin
			entryBase := entries^;
			pos := 0;
			entriesHandleSize := GetHandleSize(entries);
			while (ndx > 1) & (pos < entriesHandleSize) do begin
				entrySize := GetShort(entryBase);
				pos := pos + entrySize;
				entryBase := Ptr(ord(entryBase) + entrySize);
				ndx := ndx - 1;
			end; (* while *)
			err := ICGetMapEntry(globals^.target, entries, pos, entry);
		end; (* if *)
		ICCIGetIndMapEntry := err;
	end; (* ICCIGetIndMapEntry *)

	function ICCIGetMapEntry (globals : ICInstanceGlobalsPtr; entries: Handle; pos: longint; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entriesSize : longint;
			junkUserDataSize: longint;
	begin
		(* Check parameters. *)
		{$unused globals}
		err := noErr;

		ICParamCheck(err, entries <> nil, 'ICGetMapEntry: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'ICGetMapEntry: entries has been purged');
		ICParamCheck(err, pos >= 0, 'ICGetMapEntry: pos is negative');
		// Anarchie regularly trips the following parameter check as it walks through
		// the entries handle by calling ICGetMapEntry with pos = entriesSize.  This is
		// legal -- IC defines it will return paramErr in this case).  However,
		// tripping the DebugStr is a pain, so I've change the check to handle this case
		// explicitly.
		if err = noErr then begin
			entriesSize := GetHandleSize(entries);
			if pos = entriesSize then begin
				err := paramErr;
			end else begin
				ICParamCheck(err, pos < entriesSize, 'ICGetMapEntry: pos is off the end of entries');
			end; (* if *)
		end; (* if *)

		(* Simply call UnpackEntry to unpack the entry for the client. *)
		if err = noErr then begin
			err := UnpackEntry(entries, pos, entry, junkUserDataSize);
		end; (* if *)
		ICCIGetMapEntry := err;
	end; (* ICCIGetMapEntry *)

	function ICCISetMapEntry (globals : ICInstanceGlobalsPtr; entries: Handle; pos: longint; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			oldEntry: ICMapEntry;
			oldEntryLength: longint;
			userDataSize: longint;
			packedNewEntry: ICMapEntry;
			junkLong: longint;
	begin
		(* Check parameters. *)
		{$unused globals}
		err := noErr;
		ICParamCheck(err, entries <> nil, 'ICSetMapEntry: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'ICSetMapEntry: entries has been purged');
		ICParamCheck(err, pos >= 0, 'ICSetMapEntry: pos is negative');
		ICParamCheck(err, pos < GetHandleSize(entries), 'ICSetMapEntry: pos is off the end of entries');

		(* First we have to unpack the entry to find out the size of the user data. *)
		if err = noErr then begin
			err := UnpackEntry(entries, pos, oldEntry, userDataSize);
		end; (* if *)

		(* Now we pack the new entry into packedNewEntry, and set it up
			so that it indicates there is userDataSize bytes of user data.
		*)
		if err = noErr then begin
			PackEntry(entry, @packedNewEntry, userDataSize);
			oldEntryLength := oldEntry.total_length - userDataSize;
			
			(* If there are less than 8 bytes of user data, remove it.  This
				is basically a hack to remove alignment bytes that the previous
				version of IC used to generate.  It works because we've now
				defined a structure to the user bytes such that the minimum
				size of the user bytes is 8.
			*)
			if userDataSize < 8 then begin
				oldEntryLength := oldEntry.total_length;
				packedNewEntry.total_length := packedNewEntry.total_length - userDataSize;
				userDataSize := 0;
			end; (* if *)
			
			(* Finally call Munger to delete oldEntryLength worth of bytes
				and replace them with the new entry.  Note that we don't
				delete the user bytes (because oldEntryLength doesn't
				contain the user bytes) and hence they remain in the
				handle when we insert the new entry.
			*)
			junkLong := Munger(entries, pos,
											nil, oldEntryLength,
											@packedNewEntry, packedNewEntry.total_length - userDataSize);
			err := MemError;
 		end; (* if *)
		ICCISetMapEntry := err;
	end; (* ICCISetMapEntry *)

	function ICCIDeleteMapEntry (globals : ICInstanceGlobalsPtr; entries: Handle; pos: longint): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entry: ICMapEntry;
			junkUserDataSize: longint;
			junk: longint;
	begin
		(* Check parameters. *)
		{$unused globals}
		err := noErr;
		ICParamCheck(err, entries <> nil, 'ICDeleteMapEntry: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'ICDeleteMapEntry: entries has been purged');
		ICParamCheck(err, pos >= 0, 'ICDeleteMapEntry: pos is negative');
		ICParamCheck(err, pos < GetHandleSize(entries), 'ICDeleteMapEntry: pos is off the end of entries');

		(* Unpack the entry solely to get the total_length. *)
		if err = noErr then begin
			err := UnpackEntry(entries, pos, entry, junkUserDataSize);
		end; (* if *)
		
		(* Now delete that many bytes from the pos position in the handle. *)
		if err = noErr then begin
			junk := Munger(entries, pos, nil, entry.total_length, Ptr(-1), 0);
			err := MemError;
		end; (* if *)
		ICCIDeleteMapEntry := err;
	end; (* ICCIDeleteMapEntry *)

	function ICCIAddMapEntry (globals : ICInstanceGlobalsPtr; entries: Handle; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			newEntry: ICMapEntry;
	begin
		(* Check parameters. *)
		{$unused globals}
		err := noErr;
		ICParamCheck(err, entries <> nil, 'ICAddMapEntry: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'ICAddMapEntry: entries has been purged');

		(* Pack the passed in entry into newEntry, and then append it to the handle. *)
		if err = noErr then begin
			PackEntry(entry, @newEntry, 0);
			err := PtrAndHand(@newEntry, entries, entry.total_length);
		end; (* if *)
		ICCIAddMapEntry := err;
	end; (* ICCIAddMapEntry *)

	(* ***** Mid Level Mapping Routines ***** *)

	GLOBAL_TRACE_SETUP(gICCIMapEntriesFilename)
	
	function ICCIMapEntriesFilename (globals : ICInstanceGlobalsPtr; entries: Handle; filename: StringPtr; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		(* implementation lifted directly from Space Aliens *)
		var
			err: ICError;
			copiedFilename: Str255;
			posIndex: longint;
			longestLenSoFar: integer;
			posOfLongestLenSoFar: longint;
			TRACE_SETUP
	begin
		LOG_ENTRY(gICCIMapEntriesFilename);
		(* Check parameters. *)
		copiedFilename := '';
		err := noErr;
		if (err = noErr) and (filename <> nil) then begin
			BlockMoveData(filename, @copiedFilename, ord(filename^[0]) + 1);
		end; (* if *)
		ICParamCheck(err, entries <> nil, 'ICMapEntriesFilename: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'ICMapEntriesFilename: entries has been purged');
		ICParamCheck(err, copiedFilename <> '', 'ICMapEntriesFilename: filename must not be nil or empty');

		(* Loop through each of the entries in the handle looking for the
			longest match.
		*)
		if err = noErr then begin
			longestLenSoFar := 0;
			posIndex := 0;
			while FastGetEntry(entries, posIndex, entry) = noErr do begin
				(* We've found a new longest match if:
					o not_incoming flag bit is clear
					o it's longer than the previous max
					o it's longer than the file name
					o it matches the last N chars of the filename
				*)
				if (length(entry.extension) > longestLenSoFar) & 
							not btst(entry.flags, ICmap_not_incoming_bit) & 
									IsExtensionVar(copiedFilename, entry.extension) then begin
					(* Record the new longest entry. *)
					longestLenSoFar := length(entry.extension);
					posOfLongestLenSoFar := posIndex;
				end; (* if *)
				(* Increment posIndex so that we get the next entry the next time
					around the loop.
				*)
				posIndex := posIndex + entry.total_length;
			end; (* while *)
		end; (* if *)

		(* If we never found a match, leave with an error. *)
		if (err = noErr) & (longestLenSoFar = 0) then begin
			err := icPrefNotFoundErr;
		end; (* if *)

		(* If all is OK, return the found entry to the client. *)
		if err = noErr then begin
			err := ICGetMapEntry(globals^.target, entries, posOfLongestLenSoFar, entry);
		end; (* if *)
		LOG_EXIT(gICCIMapEntriesFilename);
		ICCIMapEntriesFilename := err;
	end; (* ICCIMapEntriesFilename *)

	function ICCIMapEntriesTypeCreator (globals : ICInstanceGlobalsPtr; entries: Handle; fType: OSType; fCreator: OSType; filename: StringPtr; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			copiedFilename: Str255;
			posIndex: longint;
			weightOfThisEntry: longint;
			bestWeightSoFar: longint;
			posOfBestWeightSoFar: longint;
	begin
		(* Check parameters. *)
		err := noErr;
		ICParamCheck(err, entries <> nil, 'ICMapEntriesTypeCreator: entries is nil');
		ICParamCheck(err, entries^ <> nil, 'ICMapEntriesTypeCreator: entries has been purged');

		(* Copy in the filename string to a local variable.  Note that we
			don't check it for the empty string because empty strings
			are legal for this routine.
		*)
		copiedFilename := '';
		if filename <> nil then begin
			BlockMoveData(filename, @copiedFilename, ord(filename^[0]) + 1);
		end; (* if *)
		
		(* Loop through each of the entries in the handle looking for the
			best match.  "Best match" is defined by:
			o entry must not be marked "not outgoing"
			o entry type must match fType parameter
			o entries with longer extensions that match are better
			o entries with matching creators are better

			This is implemented by a weighting mechanism.  Entries
			gain a weight of 1 if the creators match.  Entries gain
			a weight of 2 * length(extension) if the extensions match.
		*)
		if err = noErr then begin
			posIndex := 0;
			bestWeightSoFar := -1;
			while FastGetEntry(entries, posIndex, entry) = noErr do begin
				(* Entries marked "not outgoing" are not eligible. *)
				if not btst(entry.flags, ICmap_not_outgoing_bit) then begin
					if entry.file_type = fType then begin
						weightOfThisEntry := ord(entry.file_creator = fCreator);
						if IsExtensionVar(copiedFilename, entry.extension) then begin
							weightOfThisEntry := weightOfThisEntry + 2 * length(entry.extension);
						end; (* if *)
						if weightOfThisEntry > bestWeightSoFar then begin
							(* Record the new "best match" entry. *)
							posOfBestWeightSoFar := posIndex;
							bestWeightSoFar := weightOfThisEntry;
						end; (* if *)
					end; (* if *)
				end; (* if *)
				(* Increment posIndex so that we get the next entry the next time
					around the loop.
				*)
				posIndex := posIndex + entry.total_length;
			end; (* while *)

			(* If we never found a match, leave with an error, otherwise
				return the found entry to the client.
			*)
			if bestWeightSoFar = -1 then begin
				err := icPrefNotFoundErr;
			end else begin
				err := ICGetMapEntry(globals^.target, entries, posOfBestWeightSoFar, entry);
			end; (* if *)
		end; (* if *)
		ICCIMapEntriesTypeCreator := err;
	end; (* ICCIMapEntriesTypeCreator *)

	(* ***** High Level Mapping Routines ***** *)

	function ICCIMapFilename (globals : ICInstanceGlobalsPtr; filename: StringPtr; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entries: Handle;
			junkAttr: ICAttr;
	begin
		err := ICGetPrefHandle(globals^.target, kICMapping, junkAttr, entries);
		if err = noErr then begin
			err := ICMapEntriesFilename(globals^.target, entries, filename^, entry);
			DisposeHandle(entries);
		end; (* if *)
		ICCIMapFilename := err;
	end; (* ICCIMapFilename *)

	function ICCIMapTypeCreator (globals : ICInstanceGlobalsPtr; fType: OSType; fCreator: OSType; filename: StringPtr; var entry: ICMapEntry): ICError;
		(* See comment in ICAPI.p. *)
		var
			err: ICError;
			entries: Handle;
			junkAttr: ICAttr;
	begin
		err := ICGetPrefHandle(globals^.target, kICMapping, junkAttr, entries);
		if err = noErr then begin
			err := ICMapEntriesTypeCreator(globals^.target, entries, fType, fCreator, filename^, entry);
			DisposeHandle(entries);
		end; (* if *)
		ICCIMapTypeCreator := err;
	end; (* ICCIMapTypeCreator *)

	(* ***** Interrupt Safe API Routines ***** *)

	function SetInterrruptSafeMapping(sysGlobals : ICSystemGlobalsPtr; entries : Ptr; entriesSize : longint) : ICError;
		(*	Note that this explicitly works even if the current mappings
			global is nil.  This is useful when its called the first time
			an interrupt safe client is registered.
			
			We pass in sysGlobals rather than globals because this can be called
			by the multi-user code which doesn't have a component globals handy.
		*)
		var
			err : ICError;
			oldMappings : ICMappingsRecord;
			newMappings : Ptr;
	begin
		ICAssert(entries <> nil);
		ICAssert(entriesSize <> 0);
		
		ICWatchPoint(kICDebugInterruptSafeEventsMask, '(Re)creating interrupt safe mappings', 0);

		oldMappings.mappings := nil;
		
		{$ifc TARGET_API_MAC_CARBON}
			newMappings := NewPtr(entriesSize);
		{$elsec}
			newMappings := NewPtrSys(entriesSize);
		{$endc}
		err := CheckMemError(newMappings);
		if err = noErr then begin
			BlockMoveData(entries, newMappings, entriesSize);
			
			(*	OK, I'll admit, this is scary.  We want to atomically switch between
				the old and the new mappings.  But the mappings information is two
				longs (mappings and mappingsSize) so we can't use an atomic operation.
				Previously we disabled interrupts, but that's not going to cut it under
				Carbon.  So we now use a new strategy.
				
				Everyone who accesses the mappings fields does so through the
				currentMappingsRecord pointer, which almost always points to the
				global mappingsRecord field.  This code switches the pointer
				to point to a temporary copy of mappingsRecord (oldmappings), then
				munges the main copy, and then switches the pointer back.  This
				works because:
				
				a)	The pointer manipulation is atomic.  If an interrupt occurs,
					it will either see the existing currentMappingsRecord (pointing
					to mappingsRecord), or the new one (pointing to oldMappings),
					never a mixture of the two.
				b)	Interrupts always run to completion.  If an interrupt occurs while
					currentMappingsRecord points to oldMappings, it will complete before
					it returns.  Thus we can be assured that, once we restore
					currentMappingsRecord to point to mappingsRecord, no one will be
					holding on to a pointer to oldMappings.
				
				The second assumption is not true in the face of pre-emptive threading.
				For this and many other reasons, IC is not pre-emptive thread safe.
			*)
			
			oldMappings := sysGlobals^.mappingsRecord;
			sysGlobals^.currentMappingsRecord := @oldMappings;
			sysGlobals^.mappingsRecord.mappings := newMappings;
			sysGlobals^.mappingsRecord.mappingsSize := entriesSize;
			sysGlobals^.currentMappingsRecord := @sysGlobals^.mappingsRecord;
		end; (* if *)
		
		// Clean up.
		
		if oldMappings.mappings <> nil then begin
			DisposePtr(oldMappings.mappings);
			ICAssert(err = noErr);
		end; (* if *)
		
		SetInterrruptSafeMapping := err;
	end; (* SetInterrruptSafeMapping *)
	
	function RebuildInterruptSafeMappingSysGlobals(sysGlobals : ICSystemGlobalsPtr; inst : ICInstance) : ICError;
		var
			err : ICError;
			mappingsHandle : Handle;
			attr : ICAttr;
	begin
		
		// Get the mappings preference from IC.
		
		mappingsHandle := NewHandle(0);
		err := CheckMemError(mappingsHandle);
		if err = noErr then begin
			err := ICFindPrefHandle(inst, kICMapping, attr, mappingsHandle);
		end; (* if *)
		
		// Copy it into the system heap and set up sysGlobals to point to it.
		
		if err = noErr then begin
		
			(*	No need to preserve the state because we're about to dispose
				of this handle at the end of this routine.
			*)
			
			HLock(mappingsHandle);
			
			err := SetInterrruptSafeMapping(sysGlobals, mappingsHandle^, GetHandleSize(mappingsHandle));
		end; (* if *)
		
		// Clean up.
		
		if mappingsHandle <> nil then begin
			DisposeHandle(mappingsHandle);
			ICAssert(err = noErr);
		end; (* if *)
		
		RebuildInterruptSafeMappingSysGlobals := err;
	end; (* RebuildInterruptSafeMappingSysGlobals *)

	function RebuildInterruptSafeMapping(globals : ICInstanceGlobalsPtr) : ICError;
		var
			err : OSStatus;
	begin
		ICAssert(globals^.instanceFileSpec = @globals^.sysGlobals^.primaryConfigFile);

		err := RebuildInterruptSafeMappingSysGlobals(globals^.sysGlobals, globals^.target);
		RebuildInterruptSafeMapping := err;
	end; (* RebuildInterruptSafeMapping *)
	
	function ICCIRequiresInterruptSafe(globals : ICInstanceGlobalsPtr) : ICError;
	   	(*	Register as a client of the interrupt safe routines
	    	the only way to unregister is by closing the instance
	    	You can only do this if the instance is configured with
	    	the default preference file.  Once you do this, you can
	    	not reconfigure.
	    *)
	    var
	    	err : ICError;
	begin
		ICAssert(globals^.sysGlobals <> nil);
		
		err := noErr;
		ICParamCheck(err, globals^.instanceFileSpec = @globals^.sysGlobals^.primaryConfigFile, 'ICRequiresInterruptSafe: Must have default configuration');
		if err = noErr then begin
		
			ICAssert(globals^.sysGlobals^.currentMappingsRecord = @globals^.sysGlobals^.mappingsRecord);

			(*	If we're the first interrupt safe instance being opened,
				create the interrupt safe mapping table in the system
				heap.
			*)
			
			if globals^.sysGlobals^.interruptSafeCount = 0 then begin
				ICAssert(globals^.sysGlobals^.currentMappingsRecord^.mappings = nil);
				err := RebuildInterruptSafeMapping(globals);
			end else begin
				ICAssert(globals^.sysGlobals^.currentMappingsRecord^.mappings <> nil);
			end; (* if *)
			if err = noErr then begin
				globals^.sysGlobals^.interruptSafeCount := globals^.sysGlobals^.interruptSafeCount + 1;
				globals^.registeredAsInterruptSafe := true;
			end; (* if *)
		end; (* if *)
		ICCIRequiresInterruptSafe := err;
	end; (* ICCIRequiresInterruptSafe *)

	function ICCIGetMappingInterruptSafe(globals : ICInstanceGlobalsPtr; var mappingPref : Ptr; var mappingPrefSize : longint) : ICError;
		var
			err : ICError;
	begin
		// Check parameters.  While theoretically checking the var parameters
		// for nil is not necessary, it will probably help out some poor C
		// programmer who thinks they can pass nil to var parameters.
		
		err := noErr;
		ICParamCheck(err, globals^.registeredAsInterruptSafe, 'ICGetMappingInterruptSafe: Must be registered as interrupt safe');
		ICParamCheck(err, @mappingPref <> nil, 'ICGetMappingInterruptSafe: mappingPref must not be nil');
		ICParamCheck(err, @mappingPrefSize <> nil, 'ICGetMappingInterruptSafe: mappingPrefSize must not be nil');
		
		// Now return our global copy of the mappings database.

		if err = noErr then begin
			ICAssert(globals^.instanceFileSpec = @globals^.sysGlobals^.primaryConfigFile);
			ICAssert(globals^.sysGlobals <> nil);
			ICAssert(globals^.sysGlobals^.interruptSafeCount > 0);
			ICAssert(globals^.sysGlobals^.currentMappingsRecord^.mappings <> nil);
			mappingPref := globals^.sysGlobals^.currentMappingsRecord^.mappings;
			mappingPrefSize := globals^.sysGlobals^.currentMappingsRecord^.mappingsSize;
		end; (* if *)

		ICCIGetMappingInterruptSafe := err;
	end; (* ICCIGetMappingInterruptSafe *)
	
	function ICCIGetSeedInterruptSafe(globals : ICInstanceGlobalsPtr; var seed : longint) : ICError;
		var
			err : ICError;
	begin
		// Check parameters.
		
		err := noErr;
		ICParamCheck(err, globals^.registeredAsInterruptSafe, 'ICGetSeedInterruptSafe: Must be registered as interrupt safe');
		ICParamCheck(err, @seed <> nil, 'ICGetSeedInterruptSafe: seed must not be nil');
		
		// Now return our global copy of the mappings database.

		if err = noErr then begin
			ICAssert(globals^.instanceFileSpec = @globals^.sysGlobals^.primaryConfigFile);
			ICAssert(globals^.sysGlobals <> nil);
			ICAssert(globals^.sysGlobals^.interruptSafeCount > 0);
			seed := globals^.sysGlobals^.primarySeed;
		end; (* if *)

		ICCIGetSeedInterruptSafe := err;
	end; (* ICCIGetSeedInterruptSafe *)

	{$ifc icInstrumentation}

		procedure InitICMappingInstrumentation;
			var
				junk : ICError;
		begin
			// junk := InstCreateTraceClass(kInstRootClassRef, "InternetConfig:ICCIGetMapEntryInterruptSafe", OSType(0), kInstEnableClassMask, gICCIGetMapEntryInterruptSafe);
			// ICAssert(junk = noErr);

			// junk := InstCreateTraceClass(kInstRootClassRef, "InternetConfig:UnpackEntryInterruptSafe", OSType(0), kInstEnableClassMask, gUnpackEntryInterruptSafe);
			// ICAssert(junk = noErr);

			// junk := InstCreateTraceClass(kInstRootClassRef, "InternetConfig:CopyStringOut", OSType(0), kInstEnableClassMask, gCopyStringOut);
			// ICAssert(junk = noErr);
			
			junk := InstCreateTraceClass(kInstRootClassRef, "InternetConfig:ICCIMapEntriesFilename", OSType(0), kInstEnableClassMask, gICCIMapEntriesFilename);
			ICAssert(junk = noErr);
			
			// junk := InstCreateTraceClass(kInstRootClassRef, "InternetConfig:FastGetEntry", OSType(0), kInstEnableClassMask, gFastGetEntry);
			// ICAssert(junk = noErr);
		end; (* InitICMappingInstrumentation *)

	{$endc}
	
end. (* ICMapping *)