unit ICProfiles;

(* Start Header *)

(* File:         ICProfiles.p
 * Generated by: 1.0d8
 * For:          IC 2.1
 * On:           Monday, 4 October 1999, 0:53:14
 * 
 * This file is part of the Internet Configuration system and
 * is placed in the public domain for the benefit of all.
 *)

(* End Header *)

(* This file contains the core implementation of the Internet Config API.
	It defines the format of the ICRRecord type (which the outside world
	sees indirectly by way of the ICInstance, a pointer to this record) and
	the various IC API operations on that type.
*)

interface

	uses
		Types,
		Files, 
		QuickDraw, 
		Aliases, 
		Memory,
		Components,
		
		InternetConfig,
		ICCore;

	(* ***** Profile Utilities ***** *)
	
	function IsICxxResource(thisType : ResType) : Boolean;
		(* Returns true if thisType is a valid 'ICxx' resource
			type, ie its is 'ICxy' where 'x' and 'y' are valid
			profile letters.
		*)

	function AddDefaultPreferencesToProfile(globals : ICInstanceGlobalsPtr; newID : ICProfileID) : ICError;
		(*	Adds default preferences to the profile denoted by newID.  Note that this
			routine takes an implicit parameter, the resource file to which to add
			the preferences, in CurResFile.  Normally this would be globals^.configFileRefNum,
			but it's possible for this not to be the case when this routine is called
			by ICCIBegin.  Either way, it should always be the file denoted by
			globals^.instanceFileSpec^ that we're modifying.
		*)
	
	(*	These interfaces are automagically generated by the Internet Config
		Interface Generator.  See "InternetConfig.p" for a description of
		the behaviour of these routines.
	*)

(* Start ICCIProfiles.p *)

	function ICCIGetCurrentProfile(globals : ICInstanceGlobalsPtr; var currentID : ICProfileID) : ICError;
	
	function ICCISetCurrentProfile(globals : ICInstanceGlobalsPtr; newID : ICProfileID) : ICError;
	
	function ICCICountProfiles(globals : ICInstanceGlobalsPtr; var count : longint) : ICError;
	
	function ICCIGetIndProfile(globals : ICInstanceGlobalsPtr; index : longint; var thisID : ICProfileID) : ICError;
	
	function ICCIGetProfileName(globals : ICInstanceGlobalsPtr; thisID : ICProfileID; var name : Str255) : ICError;
	
	function ICCISetProfileName(globals : ICInstanceGlobalsPtr; thisID : ICProfileID; name : StringPtr) : ICError;
	
	function ICCIAddProfile(globals : ICInstanceGlobalsPtr; prototypeID : ICProfileID; var newID : ICProfileID) : ICError;
	
	function ICCIDeleteProfile(globals : ICInstanceGlobalsPtr; thisID : ICProfileID) : ICError;
	

(* End ICCIProfiles.p *)

implementation

	uses
		Resources,
		GestaltEqu, 
		OSUtils, 
		ToolUtils, 
		StandardFile, 
		AppleTalk, 
		Folders, 
		Processes, 
		Errors, 
		AppleEvents, 
		LowMem,
		Script,
		TextUtils,

		ICDebug,
		ICStringLists,
		ICCommonSubs,
		ICMakeUniqueName,
		ICResourceForkSanity, 
		ICCore,
		ICMapping,
		ICExtensionGlobals,
		ICExtensionResources;

	(* ***** Profile Utilities ***** *)

	function IsICxxResource(thisType : ResType) : Boolean;
	begin
		IsICxxResource := (thisType[1] = kICPrefResourceTypeBase[1]) &
							(thisType[2] = kICPrefResourceTypeBase[2]) &
							(thisType[3] >= kICPrefResourceTypeBase[3]) &
							(thisType[3] < chr(ord(kICPrefResourceTypeBase[3]) + ord(kNumberOfDigitsInThirdChar))) &
							(thisType[4] >= kICPrefResourceTypeBase[4]) &
							(thisType[4] < chr(ord(kICPrefResourceTypeBase[4]) + ord(kNumberOfDigitsInFourthChar)));
	end; (* IsICxxResource *)
	
	function BuildProfileCache(globals : ICInstanceGlobalsPtr; var cacheSize : integer) : ICError;
		var
			err : ICError;
			oldResFile : integer;
			typeCount : integer;
			typeIndex : integer;
			thisType : ResType;
	begin
		err := CheckInside(globals, icReadOnlyPerm);
		if err = noErr then begin
			if globals^.profileCache = nil then begin
				oldResFile := CurResFile;
				UseResFile(globals^.configFileRefNum);
				ICAssert(ResError = noErr);
				
				(*	Allocate the profileCache with the assumption that
					all resource types are valid 'ICxx' types.
				*)
				
				typeCount := Count1Types;
				globals^.profileCache := ProfileCacheHandle(NewHandle(typeCount * sizeof(ResType)));
				err := MemError();
				if err = noErr then begin
					ICAssert(globals^.profileCache <> nil);

					(*	Loop through the resource types, adding the valid
						'ICxx' types to the profileCache.
					*)
						
					cacheSize := 0;					
					for typeIndex := 1 to Count1Types do begin
						Get1IndType(thisType, typeIndex);
						ICAssert(ResError = noErr);
						if IsICxxResource(thisType) then begin
							cacheSize := cacheSize + 1;
							globals^.profileCache^^[cacheSize] := thisType;
						end; (* if *)
					end; (* for *)
					
					(*	cacheSize must be less than typeCount because it's
						increment at most once for each loop iteration, and there
						at most typeCount of those.
					*)
					
					ICAssert(cacheSize <= typeCount);
					
					(*	Reduce the profileCache size to hold only the valid
						elements.
					*)
					
					SetHandleSize(Handle(globals^.profileCache), cacheSize * sizeof(ResType));
					
					(*	Assert that the SetHandleSize worked.  It should always work
						(because we're making the handle smaller) so we don't handle
						the error cleanly, but we still check.
					*)
					
					ICAssert(MemError = noErr);
				
				end; (* if *)

				UseResFile(oldResFile);
				ICAssert(ResError = noErr);
			end else begin
				cacheSize := GetHandleSize(Handle(globals^.profileCache)) div sizeof(ResType);
			end; (* if *)
		end; (* if *)
		BuildProfileCache := err;
	end; (* BuildProfileCache *)
	
	function ProfileCacheSize(globals : ICInstanceGlobalsPtr) : integer;
	begin
		ICAssert(globals^.profileCache <> nil);
		ProfileCacheSize := GetHandleSize(Handle(globals^.profileCache)) div sizeof(ResType);
	end; (* ProfileCacheSize *)

	procedure BlowAwayProfileCache(globals : ICInstanceGlobalsPtr);
	begin
		ICAssert(globals^.profileCache <> nil);
		DisposeHandle(Handle(globals^.profileCache));
		ICAssert(MemError = noErr);
		globals^.profileCache := nil;
	end; (* BlowAwayProfileCache *)

	function CheckProfileInCache(globals : ICInstanceGlobalsPtr; thisID : ResType) : ICError;
		var
			err : ICError;
			cacheSize : integer;
			profileIndex : integer;
	begin
		ICAssert(globals^.profileCache <> nil);
		cacheSize := ProfileCacheSize(globals);
		err := icProfileNotFoundErr;
		for profileIndex := 1 to cacheSize do begin
			if globals^.profileCache^^[profileIndex] = thisID then begin
				err := noErr;
				leave;
			end; (* if *)
		end; (* for *)
		CheckProfileInCache := err;
	end; (* CheckProfileInCache *)

	function FindFreeProfileID(globals : ICInstanceGlobalsPtr; var newID : ICProfileID) : ICError;
		label
			found;
		var
			err : ICError;
			cacheSize : integer;
			thirdChar : char;
			fourthChar : char;
	begin
		ICAssert(globals^.profileCache <> nil);
		cacheSize := ProfileCacheSize(globals);
		if cacheSize >= kNumberOfDigitsInThirdChar * kNumberOfDigitsInFourthChar then begin
			ICAssert(cacheSize = kNumberOfDigitsInThirdChar * kNumberOfDigitsInFourthChar);
			err := icTooManyProfilesErr;
		end else begin
			err := icTooManyProfilesErr;
			
			(*	
				Iterate through all possible 'ICxx' types looking for one that is not
				in the cache.  The final assertion is that we found a free profile.
				This must be true because: a) profiles in the cache are unique (because
				they are returned by Get1IndType), b) the profiles in the cache
				all adhere to our 'ICxx' standard because we checked them before we
				added them to the cache, and c) we know that the cache is smaller
				than the maximum number of profiles.  Either way, if the assert trips,
				we know that something weird is happening.  Of course, in the non-debug
				case we just return the correct error.
			*)
			
			ResType(newID) := kICPrefResourceTypeBase;
			for thirdChar := kICPrefResourceTypeBase[3] to chr(ord(kICPrefResourceTypeBase[3]) + kNumberOfDigitsInThirdChar - 1) do begin
				for fourthChar := kICPrefResourceTypeBase[4] to chr(ord(kICPrefResourceTypeBase[4]) + kNumberOfDigitsInFourthChar - 1) do begin
					ResType(newID)[3] := thirdChar;
					ResType(newID)[4] := fourthChar;
					if CheckProfileInCache(globals, ResType(newID)) <> noErr then begin
						err := noErr;
						goto found;
					end; (* if *)
				end; (* for *)
			end; (* for *)
	
	found:
			ICAssert(err = noErr);
		end; (* if *)
		FindFreeProfileID := err;
	end; (* FindFreeProfileID *)
	
	function IsProfileNameUnique(params : MakeNameUniqueParamsPtr; const testName : Str255) : OSStatus;
		(* Callback for MakeNameUnique. *)
		var
			err : OSStatus;
			globals : ICInstanceGlobalsPtr;
			cacheSize : integer;
			profileIndex : integer;
			thisName : Str255;
	begin
		globals := ICInstanceGlobalsPtr(params^.refcon);
		ICAssert(globals <> nil);
		ICAssert(globals^.profileCache <> nil);

		cacheSize := ProfileCacheSize(globals);

		err := noErr;		
		for profileIndex := 1 to cacheSize do begin
			err := ICGetProfileName(globals^.target, ICProfileID(globals^.profileCache^^[profileIndex]), thisName);
			if err = noErr then begin
				if MakeUniqueEqualString(params^, thisName, testName) then begin
					err := 1;
				end; (* if *)
			end; (* if *)
			if err <> noErr then begin
				leave;
			end; (* if *)
		end; (* for *)

		IsProfileNameUnique := err;
	end; (* IsProfileNameUnique *)

	function UniqueNameForProfile(globals : ICInstanceGlobalsPtr; prototypeID : ICProfileID; var newProfileName : Str255) : ICError;
		var
			err : ICError;
			err2 : ICError;
			params : MakeNameUniqueParams;		(* This takes a bunch of stack space. *)
			startName : Str255;
			didOpen : Boolean;
	begin
		ICAssert(globals^.profileCache <> nil);
		
		(* Pick up all the bits of string names from the component's resource fork. *)
		
		err := StartUsingICResources(globals, didOpen);
		if err = noErr then begin
			params.copyStr := '';
			err := ICGetIndStringResource(globals, rMiscStrings, strCopy, @params.copyStr);
			if err = noErr then begin
				params.copyIStr := '';
				err := ICGetIndStringResource(globals, rMiscStrings, strCopyI, @params.copyIStr);
			end; (* if *)
			if err = noErr then begin
				startName := '';
				err := ICGetIndStringResource(globals, rMiscStrings, strDefaultSetName, @startName);
			end; (* if *)
		end; (* if *)
		err2 := StopUsingICResources(globals, didOpen);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		(*	Now decide whether to start with the prototype profile's name, or the
			default set name.
		*)
		
		if prototypeID <> kICNilProfileID then begin
			err := ICGetProfileName(globals^.target, prototypeID, startName);
		end; (* if *)

		if err = noErr then begin
			params.comparisonType := kMUCompareAsScriptStrings;
			params.script := smSystemScript;
			params.lang := GetScriptVariable(smSystemScript, smScriptLang);
			params.lengthLimit := 255;
			// params.copyStr
			// params.copyIStr
			params.uniqueFunction := IsProfileNameUnique;
			params.refcon := Ptr(globals);
			err := MakeNameUnique(startName, params, newProfileName);
		end; (* if *)
	
		UniqueNameForProfile := err;
	end; (* UniqueNameForProfile *)

	(* ***** Default Preference Creation ***** *)

	function GetOwnerName: Str255;
		(* Returns the Owner Name, as configured in the Sharing Setup control panel. *)
		const
			rOwnerNameString = -16096;
		var
			strH: StringHandle;
	begin
		strH := GetString(rOwnerNameString);
		if strH <> nil then begin
			(* Don't release it, someone else might be using it. *)
			GetOwnerName := strH^^;
		end else begin
			GetOwnerName := '';
		end; (* if *)
	end; (* GetOwnerName *)

	function AddDefaultPreferencesToProfile(globals : ICInstanceGlobalsPtr; newID : ICProfileID) : ICError;
		(*	See comments in interface part.

			Add the default preferences to the currently configured preferences
			file.  The basic algorithm is to copy across all the 'Pref' resources
			from component's resource fork to 'ICxx' resources in the preference
			file.  After that, we have to do some special case stuff, like
			selecting a default set of servers and configuring the RealName
			preference from the Owner Name.
			
			There's also a lot of jiggering around with version numbers to make sure
			that we only update preferences files that were last touched by a version
			of IC that is older than this version.
			
			Note that globals, newID and prefFileResourceRefnum are all implicit
			parameters to the nested routines defined below.
		*)
		var
			prefFileResourceRefnum: integer;

		function PrefExists(key: Str255): Boolean;
			(*	Returns true if a preference with the given key is already in
				the preferences file.
			*)
			var
				prefH: Handle;
		begin
			UseResFile(prefFileResourceRefnum);
			ICAssert(ResError = noErr);
			
			SetResLoad(false);
			prefH := Get1NamedResource(ResType(newID), key);
			SetResLoad(true);
			
			(*	We don't ReleaseResource(prefH) because it will be cleaned
				up when we close the preferences file at the end of this exercise.
			*)
			
			PrefExists := (prefH <> nil);
		end; (* PrefExists *)

		function NewSetPrefHandle (key: Str255; dataH: Handle) : ICError;
			(*	Store the data in dataH into the preferences file using key.
				You must set CurResFile to prefFileResourceRefnum
				before calling this routine.
			*)
			var
				err: ICError;
				err2 : ICError;
				junkLong: longint;
				emptyAttr: ICAttr;
		begin
			ICAssert(CurResFile = prefFileResourceRefnum);
			err := noErr;

			(*	Use Munger to add 4 bytes of empty attributes to the front
				of the resource.  This rigmarole avoids us duplicating the
				preference data *again* (we already have two copies of it
				in RAM during the Set1NamedResource).
			*)
			
			emptyAttr := 0;
			junkLong := Munger(dataH, 0, nil, 0, @emptyAttr, sizeof(emptyAttr));
			err := MemError;

			if err = noErr then begin
				
				(*	Now set the 'ICxx' resource in the preferences file. *)
				
				err := Set1NamedResource(dataH, ResType(newID), key);
				
				(*	Now use Munger to remove those 4 bytes of attributes. *)
				
				junkLong := Munger(dataH, 0, nil, 4, Ptr(-1), 0);
				err2 := MemError;
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)

			NewSetPrefHandle := err;
		end; (* NewSetPrefHandle *)

		function SetPrefStr(key: Str255; data: Str255) : ICError;
			(*	Stores the data in the string with the specified key 
				in preferences resource file.
			*)
			var
				err : ICError;
				stringH: StringHandle;
		begin
			err := noErr;
			if not PrefExists(key) then begin

				(*	PrefExists has unilaterally set CurResFile to prefFileResourceRefnum. *)

				(* Create a handle from the string. *)
				stringH := NewString(data);
				err := CheckMemError(stringH);

				(*	Set the preference based on that handle.  This approach
					is easier than trying to
					[Trying to what? -- Quinn, IC 2.1d2]
				*)
				
				if err = noErr then begin
					err := NewSetPrefHandle(key, Handle(stringH));
					DisposeHandle(Handle(stringH));
					ICAssert(MemError = noErr);
				end; (* if *)
			end; (* if *)
			SetPrefStr := err;
		end; (* SetPrefStr *)
		
		function SetDownloadFolder : ICError;
			var
				err : ICError;
				prefH : Handle;
				fss : FSSpec;
		begin
			prefH := NewHandle(0);
			err := CheckMemError(prefH);
			
			(*	Find the vRefNum and dirID of the desktop folder, then make a
				true FSSpec for it.
			*)
			
			if err = noErr then begin
				err := FindFolder(kOnSystemDisk, kDesktopFolderType, kCreateFolder, 
									fss.vRefNum, fss.parID);
			end; (* if *)
			if err = noErr then begin
				err := FSMakeFSSpec(fss.vRefNum, fss.parID, '', fss);
			end; (* if *)
			
			(*	Call our utility to make an ICFileSpecHandle out of the FSSpec. *)
			
			if err = noErr then begin
				err := FSSpecToICFileSpec(fss, ICFileSpecHandle(prefH));
			end; (* if *)
			
			(*	Set the pref in the file. *)
			
			if err = noErr then begin
				err := NewSetPrefHandle(kICDownloadFolder, prefH);
			end; (* if *)

			(*	Clean up. *)
			
			if prefH <> nil then begin
				DisposeHandle(prefH);
			end; (* if *)
			SetDownloadFolder := err;
		end; (* SetDownloadFolder *)

		function CheckServerDefaultPref(keyPrefix: Str255) : ICError;
			(*	This routine is used to set up default values for the various
				server preferences.  For example, there is a server preference
				"UMichAll", which contains a 'STR#' with all the UMich servers.
				There is also a preference 'UMichPreferred', which contains
				a 'STR ' that is the preferred UMich server.  We need to make
				sure that the 'UMichPreferred' preference is always initialised
				to something.  If it isn't, we set it to the first entry in the
				'UMichAll' string list.
				
				In the above example, keyPrefix would be 'UMich'.
				
				Note that this routine takes special care to only return
				an error in a case where we can't do the job, not when the
				job fails because there's something wrong.  For example, 
				we return an error if we run out of memory, but we don't
				return an error if the 'All' string list does not contain
				any strings.
			*)
			var
				err : ICError;
				serverListResource : Handle;
				junk : longint;
		begin
			err := noErr;
			if not PrefExists(concat(keyPrefix, 'Preferred')) then begin

				(* PrefExists has unilaterally set CurResFile to prefFileResourceRefnum. *)

				serverListResource := Get1NamedResource(ResType(newID), concat(keyPrefix, 'All'));
				if serverListResource <> nil then begin
				
					(*
						Now make a copy of the resource so we can munge it.

						I use HandToHand to copy the resource rather than detaching it because
						the resource has not yet been written to the resource file yet, so
						it's not possible to detach it.  DetachResource returns an error.
						So instead I clone the handle and proceed from there.
					*)

					err := HandToHand(serverListResource);
					if err = noErr then begin
						
						(*	Use Munger to remove the 4 bytes of attributes. *)
						
						junk := Munger(serverListResource, 0, nil, sizeof(ICAttr), Ptr(-1), 0);
						err := MemError;
						
						if err = noErr then begin
							if ValidStrList(serverListResource) & 
											(CountStrList(serverListResource) > 0) then begin
								err := SetPrefStr(concat(keyPrefix, 'Preferred'), 
													GetIndStrList(serverListResource, 1));
							end; (* if *)
						end; (* if *)
						
						DisposeHandle(serverListResource);
						ICAssert(MemError = noErr);
					end; (* if *)
				end; (* if *)
			end; (* if *)
			CheckServerDefaultPref := err;
		end; (* CheckServerDefaultPref *)

		function CopyAcrossPrefResources : ICError;
			(* Copies all the 'Pref' resources from the component
				resource fork to the preferences file resource fork.
				
				Note that this will also set up a default profile name,
				because we have a 'Pref' name=kICProfileName in the
				set of default preferences.
			*)
			var
				err : ICError;
				numberOfPrefResources: integer;
				prefResourceIndex : integer;
				prefData: Handle;
				prefName: Str255;
				junkID: integer;
				junkType: ResType;
		begin
			err := noErr;
			
			UseResFile(globals^.resourceFileRefNum);
			ICAssert(ResError = noErr);
			
			numberOfPrefResources := Count1Resources('Pref');
			ICAssert(ResError = noErr);
			
			(* For each 'Pref' resource... *)
			
			for prefResourceIndex := 1 to numberOfPrefResources do begin
			
				(* Get the resource. *)
				
				UseResFile(globals^.resourceFileRefNum);
				ICAssert(ResError = noErr);
				
				prefData := Get1IndResource('Pref', prefResourceIndex);
				err := CheckResError(prefData);
				if err = noErr then begin
				
					(* Get the key for the preference. *)
					
					GetResInfo(prefData, junkID, junkType, prefName);
					ICAssert(ResError = noErr);
					
					(* Set the preference. *)
					
					if not PrefExists(prefName) then begin
						err := NewSetPrefHandle(prefName, prefData);
					end; (* if *)
					
					(*
						Release the 'Pref' resource to prevent us clogging
						up memory with preferences that we never use again.
						This is a potential danger if some other part of IC
						has a handle to one of these resources, but I've checked
						and I don't think that's the case.
					*)
					
					ReleaseResource(prefData);
					ICAssert(ResError = noErr);
				end; (* if *)
				
				if err <> noErr then begin
					leave;
				end; (* if *)
			end; (* if *)
			CopyAcrossPrefResources := err;
		end; (* CopyAcrossPrefResources *)
		
		function UpdateProfileVersionResource : ICError;
			(*	profileVersionResource is a handle to the IC preference that describes
				the version of IC that last updated the profile specified by newID.
				This routine updates that resource to say that we were the latest
				version of the IC to update this profile.
			*)
			var
				err : ICError;
				profileVersionHandle : ICProfileVersionResourceHandle;
		begin
			UseResFile(prefFileResourceRefnum);
			ICAssert(ResError = noErr);

			(* Create a handle to hold the new profile version info. *)
			
			profileVersionHandle := ICProfileVersionResourceHandle(NewHandleClear(sizeof(ICProfileVersionResource)));
			err := CheckMemError(profileVersionHandle);
			if err = noErr then begin
			
				(* Fill out the handle and then set the profile version resource. *)
				
				profileVersionHandle^^.attr := 0;
				profileVersionHandle^^.version := globals^.sysGlobals^.extensionVersion;
				err := Set1NamedResource(Handle(profileVersionHandle), ResType(newID), kICProfileVersion);
				
				(* Clean up. *)
				
				DisposeHandle(Handle(profileVersionHandle));
				ICAssert(MemError = noErr);
			end; (* if *)
			UpdateProfileVersionResource := err;
		end; (* UpdateProfileVersionResource *)

		function UpdatePreferencesVersResource : ICError;
			(*	Updates the 'vers' ID=1 resource in the preferences file
				so that it's at least as modern as component's 'vers' ID=1.
				See the caller for more details about why we still do this.
			*)
			var
				err : ICError;
				preferenceVersResource : Handle;
				componentVersionResource : Handle;
		begin
			err := noErr;
			
			(*
				First decide whether we need to do any work by getting
				the resource in the preferences file and seeing whether
				it's took old.
			*)
			
			UseResFile(prefFileResourceRefnum);
			ICAssert(ResError = noErr);

			preferenceVersResource := Get1Resource('vers', 1);
			if (preferenceVersResource = nil) | 
							(longint(VersRecHndl(preferenceVersResource)^^.numericVersion) < longint(globals^.sysGlobals^.extensionVersion)) then begin
				
				ICWatchPoint(kICDebugSignificantEventsMask, 'Preference file version update', 0);

				(*	So we have something to do.  Go get the 'vers' ID=1
					resource from the component's resource file.
				*)

				UseResFile(globals^.resourceFileRefNum);
				ICAssert(ResError = noErr);

				componentVersionResource := Get1Resource('vers', 1);
				err := CheckResError(componentVersionResource);
				
				(*	Now switch back to the preference resource file and
					set the 'vers' ID=1 resource.
				*)
				
				UseResFile(prefFileResourceRefnum);
				ICAssert(ResError = noErr);
				
				err := Set1Resource(componentVersionResource, 'vers', 1);
				
				(*	componentVersionResource is still attached to the component's
					resource file and will be cleaned up when it's closed.
				*)
			end; (* if *)
			UpdatePreferencesVersResource := err;
		end; (* UpdatePreferencesVersResource *)

		var
			err : ICError;
			err2 : ICError;
			{$ifc icDebug}
				curResFileFSSpec : FSSpec;
			{$endc}
			didOpen : Boolean;
			profileVersionResource : Handle;
	begin
		ICAssert( (GetFSSpecGivenFileRefNum(CurResFile, curResFileFSSpec) = noErr) & EqualFSSpec(curResFileFSSpec, globals^.instanceFileSpec^) );
		
		prefFileResourceRefnum := CurResFile;

		(*
			Get the version associated with this profile.  The version is stored
			in a standard 'ICxx' resource, with 4 bytes of attributes followed
			by four bytes of NumVersion.
		*)
		
		profileVersionResource := Get1NamedResource(ResType(newID), kICProfileVersion);
		if (profileVersionResource = nil) | 
			(GetHandleSize(profileVersionResource) < sizeof(ICProfileVersionResource)) |
			(longint(ICProfileVersionResourceHandle(profileVersionResource)^^.version) < longint(globals^.sysGlobals^.extensionVersion)) then begin

			ICWatchPoint(kICDebugSignificantEventsMask, 'Profile version update', 0);
			
			DirtyThisFile(globals);
			
			err := StartUsingICResources(globals, didOpen);
			if err = noErr then begin
				err := CopyAcrossPrefResources;

				(* Then make sure that the servers have a valid default value. *)
				
				if err = noErr then begin
					err := CheckServerDefaultPref('Archie');
				end; (* if *)
				if err = noErr then begin
					err := CheckServerDefaultPref('InfoMac');
				end; (* if *)
				if err = noErr then begin
					err := CheckServerDefaultPref('UMich');
				end; (* if *)

				(*	Set up the Real Name preference from the owner name resource
					in the System file.
				*)
				
				if err = noErr then begin
					err := SetPrefStr(kICRealName, GetOwnerName);
				end; (* if *)
				
				(*
					Set the download folder preference to the desktop folder.
				
					Back out the below change.  Prior to IC 2.0, IC did not
					set up this preference in the extension, only in the
					application.  By setting this up in the extension, we
					trigger a bug in Internet Explorer 4.01.  [More accurately
					the bug is in MS Preferences Library.]  This was causing
					severe grief for the iMac "Custom Mac" bundle.  The
					simplest workaround was to back out this perfectly correct
					change until the fix to MS Preferences Library (which will
					be part of Internet Explorer 5.0) is distributed widely
					enough to prevent this inconveniencing a significant
					number of customers.
				
					This single line change is the only difference between
					IC 2.0.1 and IC 2.0.  The change will be carried forward
					into future versions of IC.
				
					-- Quinn, 22 Jul 1998, IC 2.0.2, Radar ID 2249650

				if err = noErr then begin
				    err := SetDownloadFolder;
				end;
				*)
				
				(*
					Update the 'vers' ID=1 resource in the preferences file
					to match that of the component.  This is not strictly speaking
					necessary (because we now have a version for each profile) but
					it has two advantages.  Firstly, the resource is visible to the
					user in the Finder's Get Info window.  Secondly, IC 1.4 keyed
					preference file updates off this resource, so this updated
					resource will prevent IC 1.4 from trying to update an IC 2.0
					preferences file.
				
					Note that we have to do this inside the
					Start/StopUsingICResources because, if we decide to
					update the 'vers' resource, we need the component's entire 
					'vers' resource, not just the version of the component
					(which we have cached away in our globals), but the short
					version, long version, and so on.
				*)
				
				if err = noErr then begin
					err := UpdatePreferencesVersResource;
				end; (* if *)

				err2 := StopUsingICResources(globals, didOpen);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
			
			(*	All went well while updating the preferences.  Update the
				kICProfileVersion resource to say that we did it.
			*)
			
			if err = noErr then begin
				err := UpdateProfileVersionResource;
			end; (* if *)
		end; (* if *)

		(*
			Call UpdateResFile here, just to make sure everything is locked down.
			Note that the Resource Manager can defer returning resource write errors 
			until this point, so it's important that we actually return the error
			code from UpdateResFile.
		*)

		UpdateResFile(prefFileResourceRefnum);
		err2 := ResError;
		if err = noErr then begin
			err := err2;
		end; (* if *)		
		
		AddDefaultPreferencesToProfile := err;
	end; (* AddDefaultPreferencesToProfile *)
		
	function CopyPreferencesFromProfileToProfile(globals : ICInstanceGlobalsPtr; prototypeID : ICProfileID; newID : ICProfileID) : ICError;
		var
			err : ICError;
			numberOfPrefResources: integer;
			protoResourceIndex : integer;
			protoData: Handle;
			protoName: Str255;
			junkID: integer;
			junkType: ResType;
			{$ifc icDebug}
				profileVersionResource : Handle;
			{$endc}
	begin
		{$ifc not icDebug}
			{$unused globals}
		{$endc}

		(*	Unlike AddDefaultPreferencesToProfile, CopyPreferencesFromProfileToProfile can
			only be called inside a Begin/End pair, so the following assertion is a lot cleaner.
		*)
		
		ICAssert(CurResFile = globals^.configFileRefNum);
		ICAssert(globals^.profileCache <> nil);
		ICAssert(CheckProfileInCache(globals, ResType(prototypeID)) = noErr);

		err := noErr;
		
		numberOfPrefResources := Count1Resources(ResType(prototypeID));
		ICAssert(ResError = noErr);
		
		(*	numberOfPrefResources must be greater than zero because a profile can only
			exist if it has some resources, and thus the prototype profile should not
			exist if numberOfPrefResources is 0.  This is just a general sanity check.
		*)
		
		ICAssert(numberOfPrefResources > 0);
		
		(* For each prototype resource... *)
		
		for protoResourceIndex := 1 to numberOfPrefResources do begin
		
			(*	Get the resource. *)
			
			protoData := Get1IndResource(ResType(prototypeID), protoResourceIndex);
			err := CheckResError(protoData);
			if err = noErr then begin
			
				(*	Get the key for the preference. *)
				
				GetResInfo(protoData, junkID, junkType, protoName);
				ICAssert(ResError = noErr);
				
				(*	Set the preference. *)
				
				err := Set1NamedResource(protoData, ResType(newID), protoName);

				(*	Release the prototype resource to prevent us clogging
					up memory with preferences that we never use again.
					This is a potential danger if some other part of IC
					has a handle to one of these resources, but I've checked
					and I don't think that's the case.
				*)
				
				ReleaseResource(protoData);
				ICAssert(ResError = noErr);
			end; (* if *)
			
			if err <> noErr then begin
				leave;
			end; (* if *)
		end; (* if *)
		
		(*	As we clone all the resources in the existing profile, we should automatically
			get a profile name and version.  In addition, the version resource should be
			properly formed, and be at least as modern as our component.  If neither of these
			conditions are met, we want to hear about it.
		*)
		
		{$ifc icDebug}
			if err = noErr then begin
				profileVersionResource := Get1NamedResource(ResType(newID), kICProfileVersion);
				ICAssert( (profileVersionResource <> nil) &
							(GetHandleSize(profileVersionResource) >= sizeof(ICProfileVersionResource)) &
							(longint(ICProfileVersionResourceHandle(profileVersionResource)^^.version) >= longint(globals^.sysGlobals^.extensionVersion)) );
				ICAssert( Get1NamedResource(ResType(newID), kICProfileName) <> nil );
			end; (* if *)
		{$endc}
	
		CopyPreferencesFromProfileToProfile := err;
	end; (* CopyPreferencesFromProfileToProfile *)
		
	(* ***** Profile Manipulation ***** *)
	
	function ICCIGetCurrentProfile(globals : ICInstanceGlobalsPtr; var currentID : ICProfileID) : ICError;
		var
			err : ICError;
			err2 : ICError;
			forceInsideInfo : Boolean;
	begin
		(*	Call ForceInside to ensure that configProfileResType is
			set up, then just return it as the profile ID.
		*)
		
		err := ForceInside(globals, icReadOnlyPerm, forceInsideInfo);
		if err = noErr then begin
			currentID := longint(globals^.configProfileResType);
		end; (* if *)
		err2 := ReleaseInside(globals, forceInsideInfo);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		(*	Post condition.  Either we're returning an error, or
			we're returning a valid profile ID.
		*)
		
		ICAssert((err <> noErr) | IsICxxResource(ResType(currentID)));
		
		ICCIGetCurrentProfile := err;
	end; (* ICCIGetCurrentProfile *)
	
	function ICCISetCurrentProfile(globals : ICInstanceGlobalsPtr; newID : ICProfileID) : ICError;
		var
			err : ICError;
			err2 : ICError;
			junk : ICError;
			oldResFile : integer;
			forceInsideInfo : Boolean;
			profileNameH : Handle;
			currentProfileResource : Handle;
	begin
		err := ForceInside(globals, icReadWritePerm, forceInsideInfo);
		
		if err = noErr then begin
			oldResFile := CurResFile;
			UseResFile(globals^.configFileRefNum);
			ICAssert(ResError = noErr);
			
			(* Validate destination profile ID. *)
			
			ICParamCheck(err, IsICxxResource(ResType(newID)), 'ICSetCurrentProfile: newID is not a profile ID');

			if err = noErr then begin
				profileNameH := Get1NamedResource(ResType(newID), kICProfileName);
				err := CheckResError(profileNameH);
				if err = resNotFound then begin
					err := icProfileNotFoundErr;
				end; (* if *)
			end; (* if *)
			
			(* Get the resource that holds the current profile setting. *)
			
			if err = noErr then begin
				currentProfileResource := Get1Resource('CURP', rCurrentProfileResource);
				err := CheckResError(currentProfileResource);
			end; (* if *)

			(*
				Update the current profile resource.  Other instances 
				do not have this value cached because we're inside an
				ICBegin(..., icReadWritePerm) and, by definition, we're
				the only instance allowed to be inside one of those.
				Other instances will re-fetch the 'CURP' resource when
				they next do an ICBegin.
				
				This updates also bumps the seed, and thus causes client's
				of other instances to notice the change.
				
				Note that we call UpdateResFile here because, if we're going
				to get an error, we want to know about it ASAP.
			*)
			
			if err = noErr then begin
				ResTypePtr(currentProfileResource^)^ := ResType(newID);
				ChangedResource(currentProfileResource);
				UpdateResFile(globals^.configFileRefNum);
				err := ResError();
			end; (* if *)
			
			(*	If that goes well, we switch our instance's copy of the resource. *)
			
			if err = noErr then begin
				DirtyThisFile(globals);
				globals^.configProfileResType := ResType(newID);
			end; (* if *)
			
			(*
				Once we switch profiles we must inform the interrupt safe
				code that we have done so, so it has chance to recache
				the interrupt safe copy of the mappings database.
			
				For reasons behind failure policy, check out comment
				in ICCISetPref.
			*)
			
			if err = noErr then begin
				if (globals^.instanceFileSpec = @globals^.sysGlobals^.primaryConfigFile) & 
						(globals^.sysGlobals^.interruptSafeCount > 0) then begin
					junk := RebuildInterruptSafeMapping(globals);
					if junk <> noErr then begin
						ICWatchPoint(kICDebugInterruptSafeEventsMask, 'Profile switch failed to update interrupt safe mappings', junk);
					end; (* if *)
				end; (* if *)
			end; (* if *)
			
			UseResFile(oldResFile);
			ICAssert(ResError = noErr);
		end; (* if *)

		err2 := ReleaseInside(globals, forceInsideInfo);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		
		(*	As we leave, we make sure that the next time Begin is called on this instance,
			the resource file and preference checking code is run.  This gets around the
			tricky problem posed in AddDefaultPreferencesToCurrentProfile (in ICCore.p)
			of whether we check (and potentially update) all the profiles in the file,
			or just the current one.  I chose to only update the current one, so we
			make sure that the check runs again if the profile is switched.
		*)
		
		if err = noErr then begin
			globals^.waitingForFirstBegin := true;
		end; (* if *)

		ICCISetCurrentProfile := err;
	end; (* ICCISetCurrentProfile *)
	
	function ICCICountProfiles(globals : ICInstanceGlobalsPtr; var count : longint) : ICError;
		var
			err : ICError;
			cacheSize : integer;
	begin
		err := BuildProfileCache(globals, cacheSize);
		if err = noErr then begin
			count := cacheSize;
		end; (* if *)
		ICCICountProfiles := err;
	end; (* ICCICountProfiles *)
	
	function ICCIGetIndProfile(globals : ICInstanceGlobalsPtr; n : longint; var thisID : ICProfileID) : ICError;
		var
			err : ICError;
			cacheSize : integer;
	begin
		(* Check parameters. *)
		
		err := BuildProfileCache(globals, cacheSize);
		ICParamCheck(err, n >= 1, 'ICGetIndProfile: n must be positive');
		if (err = noErr) & (n > cacheSize) then begin
			err := icProfileNotFoundErr;
		end; (* if *)

		(*	Return the n'th element in the cache. *)
		
		if err = noErr then begin
			thisID := longint(globals^.profileCache^^[n]);
		end; (* if *)
		
		ICCIGetIndProfile := err;
	end; (* ICCIGetIndProfile *)
	
	function ICCIGetProfileName(globals : ICInstanceGlobalsPtr; thisID : ICProfileID; var name : Str255) : ICError;
		var
			err : ICError;
			err2 : ICError;
			forceInsideInfo : Boolean;
			trueProfileResType : ResType;
			attr : ICAttr;
			prefSize : longint;
	begin
		(*	Check parameters.  Note that we call ForceInside to ensure
			that we have the resource file open, and hence make sure that
			our cache of the 'CURP' resource (ie globals^.configProfileResType)
			is valid, so that we can switch it before calling through to
			ICGetPref.
		*)
		
		err := ForceInside(globals, icReadOnlyPerm, forceInsideInfo);
		ICParamCheck(err, IsICxxResource(ResType(thisID)), 'ICGetProfileName: thisID is not a profile ID');
		ICParamCheck(err, @name <> nil, 'ICGetProfileName: name is nil');

		if err = noErr then begin
			(*	Temporarily 'switch' current configuration, then call
				through to ICGetPref to actually get the string.
			*)
			
			trueProfileResType := globals^.configProfileResType;
			globals^.configProfileResType := ResType(thisID);
			
			prefSize := 256;
			err := ICGetPref(globals^.target, kICProfileName, attr, @name, prefSize);
			if err <> noErr then begin
				name := '';
				err := noErr;
			end; (* if *)

			(* Clean up. *)
			
			globals^.configProfileResType := trueProfileResType;
		end; (* if *)		
		
		err2 := ReleaseInside(globals, forceInsideInfo);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		ICCIGetProfileName := err;
	end; (* ICCIGetProfileName *)
	
	function ICCISetProfileName(globals : ICInstanceGlobalsPtr; thisID : ICProfileID; name : StringPtr) : ICError;
		var
			err : ICError;
			err2 : ICError;
			forceInsideInfo : Boolean;
			trueProfileResType : ResType;
	begin
		(*	Check parameters.  Note that we call ForceInside to ensure
			that we have the resource file open, and hence make sure that
			our cache of the 'CURP' resource (ie globals^.configProfileResType)
			is valid, so that we can switch it before calling through to
			ICSetPref.
		*)
		
		err := ForceInside(globals, icReadWritePerm, forceInsideInfo);
		ICParamCheck(err, IsICxxResource(ResType(thisID)), 'ICSetProfileName: thisID is not a profile ID');
		ICParamCheck(err, name <> nil, 'ICSetProfileName: name is nil');

		if err = noErr then begin
			(*	Temporarily 'switch' current configuration, then call
				through to ICSetPref to actually set the preference.
			*)
			
			trueProfileResType := globals^.configProfileResType;
			globals^.configProfileResType := ResType(thisID);
			
			err := ICSetPref(globals^.target, kICProfileName, ICattr_no_change, Ptr(name), length(name^) + 1);

			(* Clean up. *)
			
			globals^.configProfileResType := trueProfileResType;
		end; (* if *)		
		
		err2 := ReleaseInside(globals, forceInsideInfo);
		if err = noErr then begin
			err := err2;
		end; (* if *)
		ICCISetProfileName := err;
	end; (* ICCISetProfileName *)
	
	function ICCIAddProfile(globals : ICInstanceGlobalsPtr; prototypeID : ICProfileID; var newID : ICProfileID) : ICError;
	
		procedure ReleaseAllPreferencesInProfile(thisID : ICProfileID);
			(*	Releases every resource of type thisID that may be lurking in
				memory.  We do this to prevent the new profile's resources taking
				up memory unless we actually switch to it.  This is very important
				if the client repeatedly creates new profiles without using them
				(and without calling ICEnd, which closes the resource file and
				cleans up the resources for us), such as our test program does.
			*)
			var
				numberOfPrefs : integer;
				prefIndex : integer;
				aPrefResource : Handle;
		begin
			ICAssert( CurResFile = globals^.configFileRefNum );
		
			numberOfPrefs := Count1Resources(ResType(thisID));
			ICAssert(ResError = noErr);
			
			for prefIndex := 1 to numberOfPrefs do begin
				SetResLoad(false);
				aPrefResource := Get1IndResource(ResType(thisID), prefIndex);
				SetResLoad(true);
				
				ICAssert(aPrefResource <> nil);
				if aPrefResource <> nil then begin
					ReleaseResource(aPrefResource);
					ICAssert( ResError = noErr );
				end; (* if *)
			end; (* for *)
		end; (* ReleaseAllPreferencesInProfile *)
		
		var
			err : ICError;
			cacheSize : integer;
			newProfileName : Str255;
			oldResFile : integer;
	begin
		err := CheckInside(globals, icReadWritePerm);
		ICParamCheck(err, (prototypeID = kICNilProfileID) | IsICxxResource(ResType(prototypeID)), 'ICAddProfile: prototypeID is neither kICNilProfileID nor a valid profile ID');
		
		(* Build the profile cache, then check for the profile in the cache. *)
		
		if err = noErr then begin
			err := BuildProfileCache(globals, cacheSize);
			if err = noErr then begin
				if prototypeID <> kICNilProfileID then begin
					err := CheckProfileInCache(globals, ResType(prototypeID));
				end; (* if *)
			end; (* if *)
		end; (* if *)
		
		(* Find a unique profile ID (newID) and name (newProfileName) for the new profile. *)

		if err = noErr then begin
			err := FindFreeProfileID(globals, newID);
		end; (* if *)
		if err = noErr then begin
			err := UniqueNameForProfile(globals, prototypeID, newProfileName);
		end; (* if *)

		(* If all is well, create the new profile. *)
		
		if err = noErr then begin
			oldResFile := CurResFile;
			UseResFile(globals^.configFileRefNum);
			ICAssert(ResError = noErr);

			if prototypeID = kICNilProfileID then begin
				err := AddDefaultPreferencesToProfile(globals, newID)
			end else begin
				err := CopyPreferencesFromProfileToProfile(globals, prototypeID, newID);
			end; (* if *)
			
			if err = noErr then begin
				err := ICSetProfileName(globals^.target, newID, newProfileName);
			end; (* if *)
			
			(* Flush out any changes to the resource file immediately. *)
			
			UpdateResFile(globals^.configFileRefNum);			
			if err = noErr then begin
				err := ResError;
			end; (* if *)
			
			if err = noErr then begin
				ReleaseAllPreferencesInProfile(newID);
			end; (* if *)
			
			UseResFile(oldResFile);
			ICAssert(ResError = noErr);
		end; (* if *)
				
		(*	If all goes well, blow away the profile cache so that the
			next user of it will rebuild it from scratch.
		*)
		
		if err = noErr then begin
			BlowAwayProfileCache(globals);
		end; (* if *)

		ICCIAddProfile := err;
	end; (* ICCIAddProfile *)
	
	function ICCIDeleteProfile(globals : ICInstanceGlobalsPtr; thisID : ICProfileID) : ICError;
		var
			err : ICError;
			cacheSize : integer;
			done : Boolean;
			toBeDeleted : Handle;
			oldResFile : integer;
	begin
		(*	Check parameters. *)
		
		err := CheckInside(globals, icReadWritePerm);
		ICParamCheck(err, IsICxxResource(ResType(thisID)), 'ICDeleteProfile: thisID is not a profile ID');
		ICParamCheck(err, globals^.configProfileResType <> ResType(thisID), 'ICDeleteProfile: trying to delete current profile');
		
		(*	Build the profile cache, then check for the profile in the cache. *)
		
		if err = noErr then begin
			err := BuildProfileCache(globals, cacheSize);
			if err = noErr then begin
				err := CheckProfileInCache(globals, ResType(thisID));;
			end; (* if *)
		end; (* if *)

		(* If all is well, let's go and remove all the resources in the profile. *)
		
		if err = noErr then begin
			oldResFile := CurResFile;
			UseResFile(globals^.configFileRefNum);
			ICAssert(ResError = noErr);

			done := false;
			repeat
				SetResLoad(false);
				toBeDeleted := Get1IndResource(ResType(thisID), 1);
				SetResLoad(true);
				if toBeDeleted = nil then begin
					done := true;
				end else begin
					RemoveResource(toBeDeleted);
					err := ResError;
				end; (* if *)
			until (err <> noErr) or done;

			(*	Flush out any changes to the resource file immediately. *)
			
			UpdateResFile(globals^.configFileRefNum);			
			if err = noErr then begin
				err := ResError;
			end; (* if *)
			
			UseResFile(oldResFile);
			ICAssert(ResError = noErr);
		end; (* if *)
				
		(*
			If all goes well, blow away the profile cache so that the
			next user of it will rebuild it from scratch.  Also mark
			the file as dirty (so that it's backed up); this also bumps
			the seed.
		*)
		
		if err = noErr then begin
			DirtyThisFile(globals);
			BlowAwayProfileCache(globals);
		end; (* if *)
		
		ICCIDeleteProfile := err;
	end; (* ICCIDeleteProfile *)

end. (* ICProfiles *)