unit ICAccessors;

(*	This file is part of the Internet Configuration system and is placed in the public domain for the benefit of all.

	Accessor functions for when not building under Carbon.  These should be provided
	by Apple as part of the Universal Interfaces, but a) they're not ready yet, and b)
	they're easy to write yourself.
*)

interface

	uses
		Types,
		Menus,
		Quickdraw,
		MacWindows,
		Dialogs,
		Controls,
		Lists,
		Translation,
		Navigation,
		Scrap,
		Folders;
		
	(* ***** Menus ***** *)
	
	type
		MenuItemIndex = SInt16;
		
	FUNCTION GetMenuEnable(menu: MenuRef): BOOLEAN;
	PROCEDURE EnableMenuItem(theMenu: MenuRef; item: MenuItemIndex);
	PROCEDURE DisableMenuItem(theMenu: MenuRef; item: MenuItemIndex);
	FUNCTION IsMenuItemEnabled(menu: MenuRef; item: MenuItemIndex): BOOLEAN;
	PROCEDURE CheckMenuItem(theMenu: MenuRef; item: INTEGER; checked: BOOLEAN);
	FUNCTION CountMenuItems(theMenu: MenuRef): INTEGER;

	(* ***** Quickdraw ***** *)
	
	FUNCTION GetRegionBounds(rgn: RgnHandle; VAR bounds: Rect): RectPtr;
	FUNCTION GetPortVisibleRegion(port: CGrafPtr; visRgn: RgnHandle): RgnHandle;
	FUNCTION GetPortTextFont(port: CGrafPtr): INTEGER;
	FUNCTION GetPortTextFace(port: CGrafPtr): ByteParameter;
	FUNCTION GetPortTextMode(port: CGrafPtr): INTEGER;
	FUNCTION GetPortTextSize(port: CGrafPtr): INTEGER;
	FUNCTION CreateNewPort: CGrafPtr;
	PROCEDURE DisposePort(port: CGrafPtr);
	FUNCTION GetQDGlobalsScreenBits(VAR screenBits: BitMap): BitMapPtr;

	(* ***** MacWindows ***** *)

{$ifc false}
	FUNCTION GetWindowRegion(window: WindowRef; inRegionCode: WindowRegionCode; ioWinRgn: RgnHandle): OSStatus;
{$endc}

	FUNCTION GetWindowPortBounds(window: WindowPtr; VAR bounds: Rect): RectPtr;
	FUNCTION GetWindowPort(window: WindowPtr): CGrafPtr;
	FUNCTION GetWindowFromPort(port: CGrafPtr): WindowPtr;
	FUNCTION ValidWindowRect(window: WindowPtr; {CONST}VAR bounds: Rect): OSStatus;

	(* ***** Dialogs ***** *)

	FUNCTION GetDialogPort(dialog: DialogPtr): CGrafPtr;
	FUNCTION GetDialogFromWindow(window: WindowPtr): DialogPtr;
	FUNCTION GetDialogWindow(dialog: DialogPtr): WindowPtr;
	FUNCTION GetDialogKeyboardFocusItem(dialog: DialogPtr): SInt16;
	PROCEDURE DisposeModalFilterUPP(userUPP: ModalFilterUPP);

	(* ***** Controls ***** *)

	FUNCTION GetControlPopupMenuHandle(control: ControlHandle): MenuHandle;

	(* ***** Lists ***** *)

	FUNCTION GetListVisibleCells(list: ListRef; VAR visible: ListBounds): ListBoundsPtr;
	PROCEDURE SetListSelectionFlags(list: ListRef; selectionFlags: OptionBits);
	PROCEDURE SetListRefCon(list: ListRef; refCon: SInt32);
	FUNCTION GetListPort(list: ListRef): CGrafPtr;
	FUNCTION GetListDataBounds(list: ListRef; VAR bounds: ListBounds): ListBoundsPtr;
	FUNCTION GetListCellSize(list: ListRef; VAR size: Point): PointPtr;
	FUNCTION GetListViewBounds(list: ListRef; VAR view: Rect): RectPtr;
	
	(* ***** Navigation ***** *)

	FUNCTION NewNavObjectFilterUPP(userRoutine: NavObjectFilterProcPtr): NavObjectFilterUPP;
	FUNCTION NewNavEventUPP(userRoutine: NavEventProcPtr): NavEventUPP;

	(* ***** Folders ***** *)

	FUNCTION NewFolderManagerNotificationUPP(userRoutine: FolderManagerNotificationProcPtr): FolderManagerNotificationUPP;
	PROCEDURE DisposeFolderManagerNotificationUPP(userUPP: FolderManagerNotificationUPP);

implementation

	uses
		MixedMode,
		
		ICDebug,
		ICCommonSubs;

	(* ***** Menus ***** *)

	FUNCTION GetMenuEnable(menu: MenuRef): BOOLEAN;
	BEGIN
		GetMenuEnable := btst(menu^^.enableFlags, 0);
	END; 

	PROCEDURE EnableMenuItem(theMenu: MenuRef; item: MenuItemIndex);
	BEGIN
		EnableItem(theMenu, item);
	END;
	
	PROCEDURE DisableMenuItem(theMenu: MenuRef; item: MenuItemIndex);
	BEGIN
		DisableItem(theMenu, item);
	END;

	FUNCTION IsMenuItemEnabled(menu: MenuRef; item: MenuItemIndex): BOOLEAN;
	BEGIN
		IsMenuItemEnabled := btst(menu^^.enableFlags, item);
	END;

	PROCEDURE CheckMenuItem(theMenu: MenuRef; item: INTEGER; checked: BOOLEAN);
	BEGIN
		CheckItem(theMenu, item, checked);
	END;
	
	FUNCTION CountMenuItems(theMenu: MenuRef): INTEGER;
	BEGIN
		CountMenuItems := CountMItems(theMEnu);
	END;

	(* ***** Quickdraw ***** *)
	
	FUNCTION GetRegionBounds(rgn: RgnHandle; VAR bounds: Rect): RectPtr;
	BEGIN
		bounds := rgn^^.rgnBBox;
		GetRegionBounds := @bounds;
	END; 

	FUNCTION GetPortVisibleRegion(port: CGrafPtr; visRgn: RgnHandle): RgnHandle;
	BEGIN
		CopyRgn(port^.visRgn, visRgn);
		GetPortVisibleRegion := visRgn;
	END; 

	FUNCTION GetPortTextFont(port: CGrafPtr): INTEGER;
	BEGIN
		GetPortTextFont := port^.txFont;
	END; 

	FUNCTION GetPortTextFace(port: CGrafPtr): ByteParameter;
	BEGIN
		GetPortTextFace := ByteParameter(port^.txFace);
	END; 

	FUNCTION GetPortTextMode(port: CGrafPtr): INTEGER;
	BEGIN
		GetPortTextMode := port^.txMode;
	END; 

	FUNCTION GetPortTextSize(port: CGrafPtr): INTEGER;
	BEGIN
		GetPortTextSize := port^.txSize;
	END; 

	FUNCTION CreateNewPort: CGrafPtr;
		VAR
			result : CGrafPtr;
	BEGIN
		result := CGrafPtr(NewPtr(sizeof(CGrafPort)));
		if result <> nil then begin
			OpenCPort(result);
		end; (* if *)
		CreateNewPort := result;
	END;

	PROCEDURE DisposePort(port: CGrafPtr);
	BEGIN
		DisposePtr(Ptr(port));
	END;

	FUNCTION GetQDGlobalsScreenBits(VAR screenBits: BitMap): BitMapPtr;
	BEGIN
		screenBits := qd.screenBits;
		GetQDGlobalsScreenBits := @screenBits;
	END;

	(* ***** MacWindows ***** *)

{$ifc false}
	FUNCTION GetWindowRegion(window: WindowRef; inRegionCode: WindowRegionCode; ioWinRgn: RgnHandle): OSStatus;
		var
			err : OSStatus;
			result : RgnHandle;
	BEGIN
		result := nil;
		case inRegionCode of
			kWindowStructureRgn:
				result := WindowPeek(window)^.strucRgn;
			kWindowContentRgn:
				result := WindowPeek(window)^.contRgn;
			
			otherwise
				; // do nothing
		end; (* case *)
		if result = nil then begin
			ICDebugStr('GetWindowRegion: Bogus inRegionCode');
			err := -1;
		end else begin
			CopyRgn(result, ioWinRgn);
			err := noErr;
		end; (* if *)
		GetWindowRegion := err;
	END; (* GetWindowRegion *)
{$endc}

	FUNCTION GetWindowPortBounds(window: WindowPtr; VAR bounds: Rect): RectPtr;
	BEGIN
		bounds := window^.portRect;
		GetWindowPortBounds := @bounds;
	END; 

	FUNCTION GetWindowPort(window: WindowPtr): CGrafPtr;
	BEGIN
		GetWindowPort := CGrafPtr(window);
	END; 

	FUNCTION GetWindowFromPort(port: CGrafPtr): WindowPtr;
	BEGIN
		GetWindowFromPort := WindowPtr(port);
	END; 

	FUNCTION ValidWindowRect(window: WindowPtr; {CONST}VAR bounds: Rect): OSStatus;
		VAR
			oldPort : GrafPtr;
	BEGIN
		GetPort(oldPort);
		SetPort(window);
		ValidRect(bounds);
		SetPort(oldPort);
		ValidWindowRect := noErr;
	END; 

	(* ***** Dialogs ***** *)

	FUNCTION GetDialogPort(dialog: DialogPtr): CGrafPtr;
	BEGIN
		GetDialogPort := CGrafPtr(dialog);
	END; 

	FUNCTION GetDialogFromWindow(window: WindowPtr): DialogPtr;
	BEGIN
		GetDialogFromWindow := window;
	END; 

	FUNCTION GetDialogWindow(dialog: DialogPtr): WindowPtr;
	BEGIN
		GetDialogWindow := dialog;
	END; 

	FUNCTION GetDialogKeyboardFocusItem(dialog: DialogPtr): SInt16;
	BEGIN
		GetDialogKeyboardFocusItem := DialogPeek(dialog)^.editField + 1;
	END; 

	PROCEDURE DisposeModalFilterUPP(userUPP: ModalFilterUPP);
	BEGIN
		DisposeRoutineDescriptor(userUPP);
	END;

	(* ***** Controls ***** *)

	FUNCTION GetControlPopupMenuHandle(control: ControlHandle): MenuHandle;
		TYPE
			MenuHandlePtr = ^MenuHandle;
			MenuHandleHandle = ^MenuHandlePtr;
	BEGIN
		GetControlPopupMenuHandle := MenuHandleHandle(control^^.contrlData)^^;
	END;

	(* ***** Lists ***** *)

	FUNCTION GetListVisibleCells(list: ListRef; VAR visible: ListBounds): ListBoundsPtr;
	BEGIN
		visible := list^^.visible;
		GetListVisibleCells := @visible;
	END; 
	
	PROCEDURE SetListSelectionFlags(list: ListRef; selectionFlags: OptionBits);
	BEGIN
		list^^.selFlags := selectionFlags;
	END; 

	PROCEDURE SetListRefCon(list: ListRef; refCon: SInt32);
	BEGIN
		list^^.refCon := refCon;
	END; 

	FUNCTION GetListPort(list: ListRef): CGrafPtr;
	BEGIN
		GetListPort := CGrafPtr(list^^.port);
	END; 

	FUNCTION GetListDataBounds(list: ListRef; VAR bounds: ListBounds): ListBoundsPtr;
	BEGIN
		bounds := list^^.dataBounds;
		GetListDataBounds := @bounds;
	END; 
	
	FUNCTION GetListCellSize(list: ListRef; VAR size: Point): PointPtr;
	BEGIN
		size := list^^.cellSize;
		GetListCellSize := @size;
	END; 

	FUNCTION GetListViewBounds(list: ListRef; VAR view: Rect): RectPtr;
	BEGIN
		view := list^^.rView;
		GetListViewBounds := @view;
	END; 
	
	(* ***** Navigation ***** *)

	FUNCTION NewNavObjectFilterUPP(userRoutine: NavObjectFilterProcPtr): NavObjectFilterUPP;
	BEGIN
		NewNavObjectFilterUPP := NewRoutineDescriptor(@userRoutine, uppNavObjectFilterProcInfo, GetCurrentArchitecture);
	END; 

	FUNCTION NewNavEventUPP(userRoutine: NavEventProcPtr): NavEventUPP;
	BEGIN
		NewNavEventUPP := NewRoutineDescriptor(@userRoutine, uppNavEventProcInfo, GetCurrentArchitecture);
	END; 

	(* ***** Folders ***** *)

	FUNCTION NewFolderManagerNotificationUPP(userRoutine: FolderManagerNotificationProcPtr): FolderManagerNotificationUPP;
	BEGIN
		NewFolderManagerNotificationUPP := NewRoutineDescriptor(@userRoutine, uppFolderManagerNotificationProcInfo, GetCurrentArchitecture);
	END; 
	
	PROCEDURE DisposeFolderManagerNotificationUPP(userUPP: FolderManagerNotificationUPP);
	BEGIN
		DisposeRoutineDescriptor(userUPP);
	END;

end. (* ICAccessors *)