program HelloCruelWorld;

	uses
		ICDebug,
		
		DriverServices,
		InternetConfig;

	procedure DoPreferenceIterateTest;
		const
			kIterationCount = 10;
		var
			err : OSStatus;
			err2 : OSStatus;
			inst : ICInstance;
			prefCount : longint;
			prefIndex : longint;
			prefName : Str255;
			startTime : AbsoluteTime;
			endTime : AbsoluteTime;
			usElapsed : Duration;
			count : longint;
	begin
		writeln('Iterating through all the preferences ', kIterationCount:1, ' times.');
		err := ICStart(inst, 'ICST');
		if err = noErr then begin
			err := ICFindConfigFile(inst, 0, nil);		// Have to do this to work with IC 2.0 for comparison testing.
		end; (* if *)
		if err = noErr then begin
			err := ICBegin(inst, icReadOnlyPerm);
			if err = noErr then begin
				startTime := UpTime;
				for count := 1 to kIterationCount do begin
					err := ICCountPref(inst, prefCount);
					if err = noErr then begin
						for prefIndex := 1 to prefCount do begin
							err := ICGetIndPref(inst, prefIndex, prefName);
							if err <> noErr then begin
								leave;
							end; (* if *)
						end; (* for *)
					end; (* if *)
					if err <> noErr then begin
						leave;
					end; (* if *)
				end; (* for *)
				endTime := UpTime;
				err2 := ICEnd(inst);
				if err = noErr then begin
					err := err2;
				end; (* if *)
			end; (* if *)
			err2 := ICStop(inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		if err = noErr then begin
			writeln('prefCount = ', prefCount:1);
			usElapsed := AbsoluteToDuration(SubAbsoluteFromAbsolute(endTime, startTime));
			if usElapsed < 0 then begin
				usElapsed := -usElapsed;
			end; (* if *)
			writeln('Success!  Time ', usElapsed:1, ' us');
		end else begin
			writeln('Failed with error ', err:1, '.');
		end; (* if *)
	end; (* DoPreferenceIterateTest *)

	procedure DoOverheadTest;
		const
			kIterationCount = 1000000;
		var
			err : OSStatus;
			err2 : OSStatus;
			inst : ICInstance;
			junkSeed : longint;
			startTime : AbsoluteTime;
			endTime : AbsoluteTime;
			usElapsed : Duration;
			count : longint;
	begin
		writeln('Estimating IC dispatch overhead by calling ICGetSeed ', kIterationCount:1, ' times.');
		err := ICStart(inst, 'ICST');
		if err = noErr then begin
			err := ICFindConfigFile(inst, 0, nil);		// Have to do this to work with IC 2.0 for comparison testing.
		end; (* if *)
		if err = noErr then begin
			startTime := UpTime;
			err := ICBegin(inst, icReadOnlyPerm);
			for count := 1 to kIterationCount do begin
				err := ICGetSeed(inst, junkSeed);
			end; (* for *)
			endTime := UpTime;
			err2 := ICStop(inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		if err = noErr then begin
			usElapsed := AbsoluteToDuration(SubAbsoluteFromAbsolute(endTime, startTime));
			if usElapsed < 0 then begin
				usElapsed := -usElapsed;
			end; (* if *)
			writeln('Success!  Time ', usElapsed:1, ' us');
		end else begin
			writeln('Failed with error ', err:1, '.');
		end; (* if *)
	end; (* DoOverheadTest *)
	
	procedure DoDataTest(goodWay : Boolean);
		const
			kKeysArrayCount = 200;
		type
			KeysArrayType = array [1..kKeysArrayCount] of Str255;
			KeysArrayPtr = ^KeysArrayType;
		var
			err  : OSStatus;
			err2 : OSStatus;
			dataH : Handle;
			prefH : Handle;
			keysArray : KeysArrayPtr;
			inst : ICInstance;
			haveStarted : Boolean;
			haveBegun : Boolean;
			prefCount : longint;
			prefIndex : longint;
			junkAttr : ICAttr;
			startTime : AbsoluteTime;
			endTime : AbsoluteTime;
			usElapsed : Duration;
	begin
		writeln('Getting all the preference data.');

		// Prepare for failure.
		
		dataH := nil;
		prefH := nil;
		keysArray := nil;
		haveStarted := false;
		haveBegun   := false;
		
		// Allocate memory.
		
		dataH := NewHandle(0);
		err := MemError;
		if err = noErr then begin
			prefH := NewHandle(0);
			err := MemError;
		end; (* if *)
		if err = noErr then begin
			keysArray := KeysArrayPtr(NewPtr(sizeof(KeysArrayType)));
			err := MemError;
		end; (* if *)
		
		// Create instance.
		
		if err = noErr then begin
			err := ICStart(inst, 'ICST');
			haveStarted := (err = noErr);
			if err = noErr then begin
				err := ICFindConfigFile(inst, 0, nil);		// Have to do this to work with IC 2.0 for comparison testing.
			end; (* if *)
		end; (* if *)
		
		// Set up keysArray.
		
		if err = noErr then begin
			err := ICBegin(inst, icReadOnlyPerm);
			haveBegun := (err = noErr);
		end; (* if *)
		if err = noErr then begin
			err := ICCountPref(inst, prefCount);
			if (err = noErr) & (prefCount > kKeysArrayCount) then begin
				err := -1;
			end; (* if *)
			if err = noErr then begin
				for prefIndex := 1 to prefCount do begin
					err := ICGetIndPref(inst, prefIndex, keysArray^[prefIndex]);
					if err <> noErr then begin
						leave;
					end; (* if *)
				end; (* for *)
			end; (* if *)
		end; (* if *)

		if haveBegun & not goodWay then begin
			// If were doing this the wrong way, ICEnd the instance
			// so that all the subsequent ICGetPrefs must open and close
			// the resource map individually.
			err2 := ICEnd(inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)

		// Now get the data for each preference and append it to dataH.
		
		if err = noErr then begin
			startTime := UpTime;
			for prefIndex := 1 to prefCount do begin
				err := ICFindPrefHandle(inst, keysArray^[prefIndex], junkAttr, prefH);
				if err = noErr then begin
					err := HandAndHand(prefH, dataH);
				end; (* if *)
				if err <> noErr then begin
					leave;
				end; (* if *)
			end; (* for *)
			endTime := UpTime;
		end; (* if *)

		if haveBegun & goodWay then begin
			// If were doing this the right way, ICEnd the instance now.
			err2 := ICEnd(inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)

		// Clean up.
		
		if haveStarted then begin
			err2 := ICStop(inst);
			if err = noErr then begin
				err := err2;
			end; (* if *)
		end; (* if *)
		
		if dataH <> nil then begin
			DisposeHandle(dataH);
			ICAssert(MemError = noErr);
		end; (* if *)

		if prefH <> nil then begin
			DisposeHandle(prefH);
			ICAssert(MemError = noErr);
		end; (* if *)

		if keysArray <> nil then begin
			DisposePtr(Ptr(keysArray));
			ICAssert(MemError = noErr);
		end; (* if *)

		// Display results;
		
		if err = noErr then begin
			writeln('prefCount = ', prefCount:1);
			usElapsed := AbsoluteToDuration(SubAbsoluteFromAbsolute(endTime, startTime));
			if usElapsed < 0 then begin
				usElapsed := -usElapsed;
			end; (* if *)
			writeln('Success!  Time ', usElapsed:1, ' us');
		end else begin
			writeln('Failed with error ', err:1, '.');
		end; (* if *)
	end; (* DoDataTest *)
	
	var
		commandStr : Str255;
begin
	writeln('ICSpeedTest!');
	writeln;
	writeln('Choose a test to run:');
	writeln('i)  Preference iterate test'); 
	writeln('o)  Overhead'); 
	writeln('dg) Data test, done the good way'); 
	writeln('db) Data test, done the bad way'); 
	readln(commandStr);
	if commandStr = 'i' then begin
		DoPreferenceIterateTest;
	end else if commandStr = 'o' then begin
		DoOverheadTest;
	end else if commandStr = 'dg' then begin
		DoDataTest(true);
	end else if commandStr = 'db' then begin
		DoDataTest(false);
	end else begin
		writeln('Huh?');
	end; (* if *)
	
	writeln('Done.  Press command-Q to Quit.');
end. (* HelloCruelWorld *)